import logging
import math
import statistics

import pandas as pd
import numpy as np

from pathlib import Path
from shapely.geometry import Point, LineString

from sdc_scissor.feature_extraction_api.road_geometry_calculator import RoadGeometryCalculator
from sdc_scissor.testing_api.test import Test


class RoadFeatures:
    def __init__(self):
        """
        Class representing the road features as attributes
        """
        self.direct_distance = 0
        self.road_distance = 0
        self.num_l_turns = 0
        self.num_r_turns = 0
        self.num_straights = 0
        self.median_angle = 0
        self.total_angle = 0
        self.mean_angle = 0
        self.std_angle = 0
        self.max_angle = 0
        self.min_angle = 0
        self.median_pivot_off = 0
        self.mean_pivot_off = 0
        self.std_pivot_off = 0
        self.max_pivot_off = 0
        self.min_pivot_off = 0
        self.start_time = 0
        self.end_time = 0
        self.test_duration = 0
        self.mean_road_diversity = 0
        self.full_road_diversity = 0
        self.safety = None

    def to_dict(self):
        """
        Serialize the road features to a python dictionary

        :return: A dictionary with the road features
        """
        members = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
        res = {}
        for member in members:
            res[member] = [getattr(self, member)]
        return res


class RoadSegment:
    def __init__(self):
        """
        A class representing a road segment
        """
        self.start_index = None
        self.end_index = None
        self.type = None
        self.angle = None
        self.radius = None
        self.segment_mean_diversity = None
        self.segment_total_diversity = None


class SegmentType:
    l_turn = "left_turn"
    r_turn = "right_turn"
    straight = "straight"


class FeatureExtractor:
    def __init__(self, segmentation_strategy):
        """
        A class that is responsible for extracting road features of test scenarios.

        :param segmentation_strategy: A road segmentation strategy
        """
        self.__road_features = RoadFeatures()
        self.__segments = []
        self.__road_geometry_calculator = RoadGeometryCalculator()
        self.__segmentation_strategy = segmentation_strategy

    @staticmethod
    def save_to_csv(road_features: list, out_dir: Path):
        """
        Save the road features to a csv file.

        :param road_features: List of road features
        :param out_dir: Path to store the csv file
        """
        logging.info("save_to_csv")
        dd = pd.DataFrame()
        for test_id, rf, duration in road_features:
            rf_dict = rf.to_dict()
            rf_dict["test_id"] = test_id
            rf_dict["duration"] = duration
            logging.info(rf_dict)
            rf_dd = pd.DataFrame(rf_dict)
            logging.info(rf_dd)
            dd = pd.concat([dd, rf_dd], ignore_index=True)

        logging.info(dd)
        out_path = out_dir / "road_features.csv"
        dd.to_csv(out_path)

    def extract_features(self, test: Test) -> RoadFeatures:
        """
        Input is a list of (x, y) tuples which defines the road.
        This function extract the angles and radius of segments.
        Furthermore, the statistics of angles and radius are calculated.

        :param test: A test object
        :return: A road feature object
        """
        segment_indexes_list = self.__segmentation_strategy.extract_segments(test.road_points)
        for indexes in segment_indexes_list:
            segment: RoadSegment = self.__get_road_segment_with_features(test, indexes)
            self.__segments.append(segment)

        self.__road_features = self.__get_full_road_features_from(test, self.__segments)
        return self.__road_features

    def __get_full_road_features_from(self, test: Test, segments: list[RoadSegment]):
        """
        Compute full road features based on all segments.

        :param test: Test object.
        :param segments: List of road segments.
        :return: An object containing road features as attributes.
        """
        road_features = RoadFeatures()
        road_features.test_duration = test.test_duration

        raw_feature_data = {"angles": [], "pivots": [], "diversities": []}

        for segment in segments:
            if segment.type == SegmentType.l_turn:
                road_features.num_l_turns += 1
            elif segment.type == SegmentType.r_turn:
                road_features.num_r_turns += 1
            elif segment.type == SegmentType.straight:
                road_features.num_straights += 1
            road_features.total_angle += segment.angle

            # these lists allows a simpler calculation of the statistics
            raw_feature_data["angles"].append(segment.angle)
            raw_feature_data["pivots"].append(segment.radius)
            raw_feature_data["diversities"].append(segment.segment_mean_diversity)

        road_features.mean_road_diversity = float(np.mean(raw_feature_data["diversities"]))
        road_features.full_road_diversity = float(np.sum(raw_feature_data["diversities"]))

        road_features.mean_angle = statistics.mean(raw_feature_data["angles"])
        road_features.median_angle = statistics.median(raw_feature_data["angles"])
        road_features.max_angle = max(raw_feature_data["angles"])
        road_features.min_angle = min(raw_feature_data["angles"])

        # more than 1 data point is required to calculate the standard deviation
        if len(raw_feature_data["angles"]) > 1:
            road_features.std_angle = statistics.stdev(raw_feature_data["angles"])
        else:
            road_features.std_angle = 0

        road_features.mean_pivot_off = statistics.mean(raw_feature_data["pivots"])
        road_features.median_pivot_off = statistics.median(raw_feature_data["pivots"])
        road_features.max_pivot_off = max(raw_feature_data["pivots"])
        road_features.min_pivot_off = min(raw_feature_data["pivots"])

        if len(raw_feature_data["pivots"]) > 1:
            road_features.std_pivot_off = statistics.stdev(raw_feature_data["pivots"])
        else:
            road_features.std_pivot_off = 0

        road_features.direct_distance = self.__road_geometry_calculator.get_distance_between(
            test.road_points[0], test.road_points[-1]
        )
        road_features.road_distance = self.__road_geometry_calculator.get_road_length(test.road_points)

        return road_features

    def __get_segment_type(self, test: Test, road_segment, angle_threshold):
        """
        Return the type of segment (straight, left turn, right turn). The segment
        is defined by its start and end index that are already specified.
        :param test:
        :param road_segment:
        :param angle_threshold:
        :return:
        """
        start_index = road_segment.start_index
        end_index = road_segment.end_index

        segment_road_points = test.road_points[start_index : end_index + 1]

        angles_lst = self.__road_geometry_calculator.extract_turn_angles(segment_road_points)

        angles_sum = sum(angles_lst)

        if angle_threshold > angles_sum > -angle_threshold:
            return SegmentType.straight
        if angles_sum >= angle_threshold:
            return SegmentType.l_turn
        if angles_sum <= angle_threshold:
            return SegmentType.r_turn

    def __get_segment_angle(self, test: Test, road_segment):
        """

        :param test:
        :param road_segment:
        :return:
        """
        start_index = road_segment.start_index
        end_index = road_segment.end_index

        segment_road_points = test.road_points[start_index : end_index + 1]

        angles_lst = self.__road_geometry_calculator.extract_turn_angles(segment_road_points)

        angles_sum = sum(angles_lst)

        return angles_sum

    def __get_segment_radius(self, test: Test, road_segment):
        """

        :param test:
        :param road_segment:
        :return:
        """
        if road_segment.type == SegmentType.straight:
            return 0

        start_index = road_segment.start_index
        end_index = road_segment.end_index

        segment_road_points = test.road_points[start_index : end_index + 1]

        angles_lst = self.__road_geometry_calculator.extract_turn_angles(segment_road_points)
        segment_length = self.__road_geometry_calculator.get_road_length(segment_road_points)

        angles_sum = sum(angles_lst)

        unsigned_angle = abs(angles_sum)

        # C=2*pi*r
        proportion = unsigned_angle / 360
        circle_length = segment_length / proportion
        radius = circle_length / (2 * math.pi)

        return radius

    def __get_segment_diversity(self, test: Test, road_segment: RoadSegment) -> tuple:
        """
        Compute the diversity of the road segment compared to a straight trajectory.

        :param test: A test object
        :param road_segment: A road segment object
        :return: The measure of diversity
        """
        road_points = test.road_points
        start_index, end_index = road_segment.start_index, road_segment.end_index
        segment_road_points = road_points[start_index : end_index + 1]
        start_point, end_point = Point(segment_road_points[0][:2]), Point(segment_road_points[-1][:2])
        direct_segment_line: LineString = LineString([start_point, end_point])

        shapely_points: list = [Point(rp[0], rp[1]) for rp in segment_road_points]
        segment_line: LineString = LineString(shapely_points)

        point_to_line_distances = [direct_segment_line.distance(point) for point in shapely_points]
        segment_total_diversity = float(np.sum(point_to_line_distances))
        segment_mean_diversity = float(np.mean(point_to_line_distances))
        return segment_total_diversity, segment_mean_diversity

    def __get_road_segment_with_features(self, test: Test, indexes) -> RoadSegment:
        """
        Compute segment features and create a segment object accordingly.

        :param test: A test object
        :param indexes: Start and end index of a single segment of the road specified in the test.
        :return: A road segment object with segment specific features.
        """
        road_segment = RoadSegment()
        road_segment.start_index = indexes[0]
        road_segment.end_index = indexes[1]

        # classify segment type
        road_segment.type = self.__get_segment_type(test, road_segment, angle_threshold=5)

        # update angle
        road_segment.angle = self.__get_segment_angle(test, road_segment)

        # calculate radius
        road_segment.radius = self.__get_segment_radius(test, road_segment)

        road_segment.segment_total_diversity, road_segment.segment_mean_diversity = self.__get_segment_diversity(
            test, road_segment
        )

        return road_segment



import numpy as np

from sdc_scissor.testing_api.test_generators.ambiegen.Utils.car_road import Map

import os
import json


class RoadGen:
    """
    Class for generating roads

    We using a Markov chain to generate a sequence of road types.
    It allows to create a better initial population

    """

    def __init__(
        self,
        map_size,
        min_len,  # minimal possible distance in meters
        max_len,  # maximal possible disance to go straight in meters
        min_angle,  # minimal angle of rotation in degrees
        max_angle,  # maximal angle of rotation in degrees
    ):
        self.file = "roads.json"
        self.init_file = "init.json"
        self.points_file = "points.json"
        self.road_points = []
        self.init_states = ["straight", "left", "right"]

        self.map_size = map_size

        self.transitionName = [["SS", "SL", "SR"], ["LS", "LL", "LR"], ["RS", "RL", "RR"]]

        self.transitionMatrix = [
            [0.1, 0.45, 0.45],
            [0.2, 0.4, 0.4],
            [0.2, 0.4, 0.4],
        ]  # probabilities of switching states

        self.min_len = min_len
        self.max_len = max_len
        self.step_ang = 5
        self.step_len = 1

        self.min_angle = min_angle
        self.max_angle = max_angle

        self.len_values = [
            i for i in range(self.min_len, self.max_len + 1, self.step_len)
        ]  # a list of distance to go forward
        self.ang_values = [
            i for i in range(self.min_angle, self.max_angle + 1, self.step_ang)
        ]  # a list of angles to turn

    def test_case_generate(self):
        """Function that produces a list with states and road points"""

        # initialization

        self.road_points = []

        self.init_states = ["straight", "left", "right"]

        self.car_map = Map(self.map_size)
        self.init_a = [int(self.car_map.init_pos[0]), int(self.car_map.init_pos[1])]
        self.init_b = [int(self.car_map.init_end[0]), int(self.car_map.init_end[1])]

        self.road_points.append(tuple(((self.init_a[0] + self.init_b[0]) / 2, (self.init_a[1] + self.init_b[1]) / 2)))

        state = np.random.choice(self.init_states)
        if state == "straight":
            value = np.random.choice(self.len_values)
        else:
            value = np.random.choice(self.ang_values)

        if state == "straight":
            self.car_map.go_straight(value)
            self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
        elif state == "left":
            self.car_map.turn_left(value)
            self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
        else:
            self.car_map.turn_right(value)
            self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))

        self.states = [[state, value]]

        flag = True

        while flag == True:
            if state == "straight":
                change = np.random.choice(self.transitionName[0], p=self.transitionMatrix[0])  # choose the next state
                if change == "SS":  # stay in the same state
                    value = np.random.choice(self.len_values)
                    state = "straight"
                    self.states.append([state, value])
                    flag = self.car_map.go_straight(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                elif change == "SL":  # change from go straight to turn left
                    value = np.random.choice(self.ang_values)
                    state = "left"
                    self.states.append([state, value])
                    flag = self.car_map.turn_left(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                elif change == "SR":
                    value = np.random.choice(self.ang_values)
                    state = "right"
                    self.states.append([state, value])
                    flag = self.car_map.turn_right(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                else:
                    print("Error")

            elif state == "left":
                change = np.random.choice(self.transitionName[1], p=self.transitionMatrix[1])
                if change == "LS":
                    value = np.random.choice(self.len_values)
                    state = "straight"
                    self.states.append([state, value])
                    flag = self.car_map.go_straight(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass

                elif change == "LL":
                    value = np.random.choice(self.ang_values)
                    state = "left"
                    self.states.append([state, value])
                    flag = self.car_map.turn_left(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass
                elif change == "LR":
                    value = np.random.choice(self.ang_values)
                    state = "right"
                    self.states.append([state, value])
                    flag = self.car_map.turn_right(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass
                else:
                    print("Error")
                    pass
            elif state == "right":
                change = np.random.choice(self.transitionName[2], p=self.transitionMatrix[2])
                if change == "RS":
                    value = np.random.choice(self.len_values)
                    state = "straight"
                    self.states.append([state, value])
                    flag = self.car_map.go_straight(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass
                elif change == "RL":
                    value = np.random.choice(self.ang_values)
                    state = "left"
                    self.states.append([state, value])
                    flag = self.car_map.turn_left(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass
                elif change == "RR":
                    value = np.random.choice(self.ang_values)
                    state = "right"
                    self.states.append([state, value])
                    flag = self.car_map.turn_right(value)
                    if flag == False:
                        del self.road_points[-1]
                        del self.states[-1]
                        if len(self.road_points) <= 2:
                            self.car_map.go_straight(1)
                            self.road_points.append(
                                tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2)
                            )
                        return self.states_to_dict()
                    self.road_points.append(tuple((self.car_map.current_pos[0] + self.car_map.current_pos[1]) / 2))
                    pass
                else:
                    print("Error")
        del self.road_points[-1]  # last point might be going over the border
        del self.states[-1]
        return self.states_to_dict()

    def states_to_dict(self):
        """Transforms a list of test cases
        to a dictionary"""
        test_cases = {}
        i = 0
        for element in self.states:
            test_cases["st" + str(i)] = {}
            test_cases["st" + str(i)]["state"] = element[0]
            test_cases["st" + str(i)]["value"] = int(element[1])
            i += 1

        return test_cases

    def write_states_to_file(self):
        """Writes the generated test case to file"""
        if os.stat(self.file).st_size == 0:
            test_cases = {}
        else:
            with open(self.file) as file:
                test_cases = json.load(file)

        if os.stat(self.init_file).st_size == 0:
            positions = {}
        else:
            with open(self.init_file) as file:
                positions = json.load(file)

        if os.stat(self.points_file).st_size == 0:
            points = {}
        else:
            with open(self.points_file) as file:
                points = json.load(file)

        num = len(test_cases)

        tc = "tc" + str(num)
        test_cases[tc] = {}
        positions[tc] = {"a": self.init_a, "b": self.init_b}
        points[tc] = self.road_points

        i = 0
        for element in self.states:
            test_cases[tc]["st" + str(i)] = {}
            test_cases[tc]["st" + str(i)]["state"] = str(element[0])
            test_cases[tc]["st" + str(i)]["value"] = int(element[1])
            i += 1

        with open(self.file, "w") as outfile:
            json.dump(test_cases, outfile)

        with open(self.init_file, "w") as outfile:
            json.dump(positions, outfile)

        with open(self.points_file, "w") as outfile:
            json.dump(points, outfile)



import logging

from sdc_scissor.feature_extraction_api.segmentation_strategy import SegmentationStrategy
from sdc_scissor.feature_extraction_api.road_geometry_calculator import RoadGeometryCalculator


class AngleBasedStrategy(SegmentationStrategy):
    """
    Define segments based on identified turns. The center of a turn where the
    angles possibly reaches the maximum will also be the center of the segment.
    Straight segments are identified where no significant turn angles are present.
    """

    def __init__(self, angle_threshold=5, decision_distance=10):
        """
        Instantiate a strategy object defining the segmentation strategy of the road. Class is coherent to the strategy pattern.

        :param angle_threshold: Angle threshold defining a new segment.
        :param decision_distance: Road distance for calculating the turn angle.
        """
        self.__road_geometry_calculator = RoadGeometryCalculator()
        self.__angle_threshold = angle_threshold
        self.__decision_distance = decision_distance

    def extract_segments(self, road_points):
        """
        Extract segments from road points according to the angle-based segmentation strategy.

        :param road_points: List of 2D  points, defining the road to be segmented
        :return: List of indexes defining the start and end of segments
        """
        # iterate according to the decision distance
        segment_indexes = []
        segment_start_index = 0
        current_road_piece_start_index = 0
        current_angle = 0
        is_first_piece = True
        is_last_iteration = False

        for i in range(len(road_points)):
            # check if it is the last iteration
            if i == (len(road_points) - 1):
                is_last_iteration = True

            # the start of a new piece has to be 2 indexes ahead of the current i
            if current_road_piece_start_index == i + 1:
                current_road_piece_end_index = i + 2
            else:
                current_road_piece_end_index = i + 1

            # define the current road piece to calculate the turn angle and distance
            current_road_piece = road_points[current_road_piece_start_index : current_road_piece_end_index + 1]

            # calculate the road piece distance defined above
            current_distance = self.__road_geometry_calculator.get_road_length(current_road_piece)

            # check if the distance of the current road piece is long enough or it is the last iteration
            if (current_distance >= self.__decision_distance) or is_last_iteration:
                previous_angle = current_angle

                # calculate the angle of the current road piece
                # TODO: ensure that the road piece to calculate the angle has enough points!!!
                # (e.g., use temporarily a longer road piece)
                if len(current_road_piece) == 2 and current_road_piece_start_index > 0:
                    tmp_current_road_piece = road_points[
                        current_road_piece_start_index - 1 : current_road_piece_end_index + 1
                    ]
                else:
                    tmp_current_road_piece = current_road_piece
                current_angle = sum(self.__road_geometry_calculator.extract_turn_angles(tmp_current_road_piece))

                # define start and end index of segment iff the angle of the current road piece is different
                if (
                    self.__has_current_angle_changed(previous_angle, current_angle) and not is_first_piece
                ) or is_last_iteration:
                    segment_end_index = i
                    segment_indexes.append((segment_start_index, segment_end_index))
                    segment_start_index = segment_end_index + 1
                    current_road_piece_start_index = segment_start_index
                else:
                    current_road_piece_start_index = current_road_piece_end_index + 1
                    is_first_piece = False

        return segment_indexes

        # decide type of segment based on angle and its threshold
        # define segment when the type of segment will change

    def __has_current_angle_changed(self, previous_angle, current_angle):
        """
        Check if there is a significant change of turn angles.

        :param previous_angle: Previous angle
        :param current_angle: Current angle to be compared with the previous angle
        :return: True or False if there is a change greater than the specified threshold
        """
        angle_threshold = self.__angle_threshold
        if current_angle <= previous_angle + angle_threshold and current_angle >= previous_angle - angle_threshold:
            return False
        return True


if __name__ == "__main__":
    logging.info("angle_based_strategy.py")



import abc


class SegmentationStrategy(abc.ABC):
    """
    Interface for implementing different segmentation strategies.
    """

    @abc.abstractmethod
    def extract_segments(self, road_points):
        """
        Abstract method to retrieve segments from a road, based on a specific strategy.

        :param road_points: Sequence of coordinates specifying the road.
        """
        pass



import logging

from sdc_scissor.simulator_api.abstract_simulator import AbstractSimulator
from sdc_scissor.simulator_api.beamng_simulator import BeamNGSimulator


class SimulatorFactory:
    @staticmethod
    def get_beamng_simulator(home, user, rf: float, max_speed: float, fov: int) -> AbstractSimulator:
        """

        :param home:
        :param user:
        :param rf:
        :param max_speed:
        :param fov: The field of view  for a Camera e.g., 120
        :return:
        """
        beamng_simulator = BeamNGSimulator(
            host="localhost", port=64256, home=home, user=user, rf=rf, max_speed=max_speed, fov=fov
        )
        return beamng_simulator


if __name__ == "__main__":
    logging.info("* simulator_factory.py")



from pymoo.model.duplicate import ElementwiseDuplicateElimination


class DuplicateElimination(ElementwiseDuplicateElimination):
    """
    Module to remove the same individuals
    """

    def is_equal(self, a, b):
        return a.X[0].states == b.X[0].states  # remove individuals that are the same



import numpy as np
from pymoo.model.mutation import Mutation
import copy
import sdc_scissor.testing_api.test_generators.ambiegen.config as cf


class TestCaseMutation(Mutation):
    """
    Module to perform the mutation
    """

    def __init__(self, mut_rate):
        super().__init__()
        self.mut_rate = mut_rate

    def _do(self, problem, X, **kwargs):
        for i in range(len(X)):
            r = np.random.random()
            s = X[i, 0]
            # with a probabilty of 40% - change the order of characters
            if r < self.mut_rate:  # cf.ga["mut_rate"]:

                sn = copy.deepcopy(s)

                sn.get_points()
                sn.remove_invalid_cases()

                wr = np.random.random()
                child = sn.states
                old_points = sn.road_points
                old_states = child
                if wr < 0.2:

                    candidates = list(np.random.randint(0, high=len(child), size=2))
                    temp = child["st" + str(candidates[0])]
                    child["st" + str(candidates[0])] = child["st" + str(candidates[1])]
                    child["st" + str(candidates[1])] = temp
                elif wr >= 0.2 and wr < 0.5:
                    num = np.random.randint(0, high=len(child))
                    value = np.random.choice(["state", "value"])
                    if value == "value":
                        duration_list = []
                        if child["st" + str(num)]["state"] == "straight":
                            duration_list = np.arange(cf.model["min_len"], cf.model["max_len"], 1)
                        else:
                            duration_list = np.arange(cf.model["min_angle"], cf.model["max_angle"], 5)

                        child["st" + str(num)][value] = int(np.random.choice(duration_list))

                    elif value == "state":

                        if child["st" + str(num)][value] == "straight":
                            child["st" + str(num)][value] = np.random.choice(["left", "right"])
                            duration_list = np.arange(cf.model["min_angle"], cf.model["max_angle"], 5)
                            child["st" + str(num)]["value"] = int(np.random.choice(duration_list))
                        else:
                            child["st" + str(num)][value] = "straight"
                            duration_list = np.arange(cf.model["min_len"], cf.model["max_len"], 1)
                            child["st" + str(num)]["value"] = int(np.random.choice(duration_list))

                else:
                    cand = list(np.random.randint(0, high=len(child), size=int(len(child) / 2)))
                    while cand:
                        c1 = np.random.choice(cand)
                        cand.remove(c1)
                        if cand:
                            c2 = np.random.choice(cand)
                            cand.remove(c2)
                            temp = child["st" + str(c1)]
                            child["st" + str(c1)] = child["st" + str(c2)]
                            child["st" + str(c2)] = temp
                        else:
                            if child["st" + str(c1)]["state"] == "straight":
                                duration_list = np.arange(cf.model["min_len"], cf.model["max_len"], 1)
                            else:
                                duration_list = np.arange(cf.model["min_angle"], cf.model["max_angle"], 5)
                            child["st" + str(c1)]["value"] = int(np.random.choice(duration_list))

                sn.states = child
                sn.get_points()
                sn.remove_invalid_cases()
                sn.novelty = sn.calc_novelty(old_states, sn.states)

                X[i, 0] = sn

        return X



import logging
import math


class RoadGeometryCalculator:
    def extract_turn_angles(self, road_points):
        """
        Extract angles of raod points and ad them to the instance variable.

        :param road_points: Points that define the road in the test scenario.
        :return: Angles in degrees of road's turns defined by the road points.
        """
        angles = []
        # iterate over "all" road points
        for i in range(2, len(road_points)):
            # calculate angle between previous direction and vector from
            # previous point to the current one

            point_before = road_points[i - 2]
            mid_point = road_points[i - 1]
            point_after = road_points[i]

            prev_direction = self.get_direction(point_before, mid_point)
            current_direction = self.get_direction(mid_point, point_after)

            turn_angle = self.get_angle(prev_direction, current_direction)
            angles.append(turn_angle)

        return angles

    @staticmethod
    def get_angle(first_vec, second_vec):
        """
        Returns the angle in degrees between the first and second vector.
        A left turn as positive angles whereas right turns have negatives.

        :param first_vec: First 2D vector
        :param second_vec: Second 3D vector
        :return: Angle between the vectors in degrees
        """
        a1, a2 = first_vec[0], first_vec[1]
        b1, b2 = second_vec[0], second_vec[1]

        angle_in_radians = math.atan2(b2, b1) - math.atan2(a2, a1)
        angle_in_degrees = math.degrees(angle_in_radians)

        return angle_in_degrees

    @staticmethod
    def get_distance_between(first_point, second_point) -> float:
        """
        Calculate Euclidean distance between two points.

        :param first_point: (x,y) coordinates of first point
        :param second_point: (x,y) coordinates of second point
        :return: Euclidean distance between the points
        """
        a1, a2 = first_point[0], first_point[1]
        b1, b2 = second_point[0], second_point[1]
        c1, c2 = (b1 - a1, b2 - a2)
        distance = math.sqrt(c1**2 + c2**2)
        return distance

    @staticmethod
    def get_direction(first_point, second_point):
        """
        Get the direction vector from the first point to the second point.

        :param first_point: (x,y) coordinates of first point
        :param second_point: (x,y) coordinates of second point
        :return: Return the difference 2D vector (second_point-first_point)
        """
        return second_point[0] - first_point[0], second_point[1] - first_point[1]

    @staticmethod
    def get_road_length(road_points):
        """
        Compute the length of the road.

        :param road_points: List of road points defining the road
        :return: Length of the road
        """
        nr_of_road_points = len(road_points)

        road_length = 0
        for i in range(1, nr_of_road_points):
            a1, a2 = road_points[i - 1][0], road_points[i - 1][1]
            b1, b2 = road_points[i][0], road_points[i][1]

            c1, c2 = (b1 - a1, b2 - a2)

            road_length += math.sqrt(c1**2 + c2**2)

        return road_length


if __name__ == "__main__":
    logging.info("road_geometry_calculator.py")



import logging
import joblib

from pathlib import Path

import numpy as np
import pandas as pd

from sdc_scissor.feature_extraction_api.feature_extraction import FeatureExtractor, RoadFeatures
from sdc_scissor.testing_api.test_loader import TestLoader


class Predictor:
    def __init__(self, test_loader: TestLoader, joblib_classifier: Path, label="safety"):
        """

        :param test_loader:
        :param joblib_classifier:
        :param label:
        """
        self.test_loader = test_loader
        self.__classifier = joblib.load(joblib_classifier)
        self.feature_extractor = FeatureExtractor(segmentation_strategy="angle-based")
        self.label = label
        self.X_model_attributes = [
            "direct_distance",
            "max_angle",
            "max_pivot_off",
            "mean_angle",
            "mean_pivot_off",
            "median_angle",
            "median_pivot_off",
            "min_angle",
            "min_pivot_off",
            "num_l_turns",
            "num_r_turns",
            "num_straights",
            "road_distance",
            "std_angle",
            "std_pivot_off",
            "total_angle",
        ]

    def predict(self):
        """
        Predict the outcome of the tests and saves is as a json property to the test file.
        """
        logging.info("predict")
        while self.test_loader.has_next():
            test, test_path = self.test_loader.next()
            road_features: RoadFeatures = self.feature_extractor.extract_features(test)
            road_features_dict = road_features.to_dict()
            road_features_df = pd.DataFrame(road_features_dict)
            X: np.ndarray = road_features_df[self.X_model_attributes].to_numpy()
            y_pred: int = self.__classifier.predict(X)[0]
            if y_pred == 1:
                test.predicted_test_outcome = "FAIL"
            elif y_pred == 0:
                test.predicted_test_outcome = "PASS"
            else:
                logging.warning("Prediction failed!")
                raise Exception("Prediction failed!")
            logging.info("predicted outcome: {}".format(test.predicted_test_outcome))
            test.save_as_json(file_path=test_path)


if __name__ == "__main__":
    logging.info("* predictor.py")



from sdc_scissor.testing_api.test_generators.frenetic.src.generators.base_generator import BaseGenerator
import sdc_scissor.testing_api.test_generators.frenetic.src.utils.frenet as frenet
import numpy as np
import logging as log


class BaseFrenetGenerator(BaseGenerator):
    def __init__(self, time_budget=None, executor=None, map_size=None, margin=10, strict_father=False):
        # Margin size w.r.t the map
        self.margin = margin
        self.recent_count = 0
        super().__init__(time_budget=time_budget, executor=executor, map_size=map_size, strict_father=strict_father)

    def kappas_to_road_points(self, kappas, frenet_step=10, theta0=1.57):
        """
        Args:
            kappas: list of kappa values
            frenet_step: The distance between to points.
            theta0: The initial angle of the line. (1.57 == 90 degrees)
        Returns:
            road points in cartesian coordinates
        """
        # Using the bottom center of the map.
        y0 = self.margin
        x0 = self.map_size / 2
        ss = np.arange(y0, (len(kappas) * frenet_step), frenet_step)

        # Transforming the frenet points to cartesian
        (xs, ys) = frenet.frenet_to_cartesian(x0, y0, theta0, ss, kappas)
        road_points = self.reframe_road(xs, ys)
        return road_points

    def execute_frenet_test(self, kappas, method="random", frenet_step=10, theta0=1.57, parent_info={}, extra_info={}):
        extra_info["kappas"] = kappas
        road_points = self.kappas_to_road_points(kappas, frenet_step=frenet_step, theta0=theta0)
        return road_points
        # if road_points:
        #     self.recent_count += 1
        #     return self.execute_test(road_points, method=method, parent_info=parent_info, extra_info=extra_info)
        # else:
        #     return 'CANNOT_REFRAME', None

    def reframe_road(self, xs, ys):
        """
        Args:
            xs: cartesian x coordinates
            ys: cartesian y coordinates
        Returns:
            A representation of the road that fits the map size (when possible).
        """
        min_xs = min(xs)
        min_ys = min(ys)
        road_width = self.margin  # TODO: How to get the exact road width?
        if (max(xs) - min_xs + road_width > self.map_size - self.margin) or (
            max(ys) - min_ys + road_width > self.map_size - self.margin
        ):
            log.info("Skip: Road won't fit")
            return None
            # TODO: Fail the entire test and start over
        xs = list(map(lambda x: x - min_xs + road_width, xs))
        ys = list(map(lambda y: y - min_ys + road_width, ys))
        return list(zip(xs, ys))



from sdc_scissor.testing_api.test_generators.ambiegen.Utils.vehicle import Car
import sdc_scissor.testing_api.test_generators.ambiegen.config as cf
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.car_road import Map

# from code_pipeline.beamng_executor import BeamngExecutor
# from code_pipeline.tests_generation import RoadTestFactory


class Solution:

    """
    This is a class to represent one individual of the genetic algorithm
    """

    def __init__(self):

        self.road_points = []
        self.states = {}
        self.car = Car(cf.model["speed"], cf.model["steer_ang"], cf.model["map_size"])
        self.road_builder = Map(cf.model["map_size"])
        self.fitness = 0
        self.car_path = []
        self.novelty = 0
        self.intp_points = []
        self.too_sharp = 0
        self.just_fitness = 0

    def eval_fitness(self):
        road = self.road_points
        if not road:  # if no road points were calculated yet
            self.get_points()
            self.remove_invalid_cases()
            road = self.road_points

        if len(self.road_points) <= 2:
            self.fitness = 0
        else:
            self.intp_points = self.car.interpolate_road(road)
            self.fitness, self.car_path = self.car.execute_road(self.intp_points)

        return

    # def car_model_fit(self):

    # the_executor = BeamngExecutor(cf.model["map_size"])

    # the_test = RoadTestFactory.create_road_test(self.road_points)

    # fit = the_executor._eval_tc(the_test)

    # return fit

    def get_points(self):
        self.road_points = self.road_builder.get_points_from_states(self.states)

    def remove_invalid_cases(self):
        self.states, self.road_points = self.road_builder.remove_invalid_cases(self.road_points, self.states)

    def calc_novelty(self, old, new):
        novelty = 0
        difference = abs(len(new) - len(old)) / 2
        novelty += difference
        if len(new) <= len(old):
            shorter = new
        else:
            shorter = old
        for tc in shorter:
            if old[tc]["state"] == new[tc]["state"]:
                value_list = [old[tc]["value"], new[tc]["value"]]
                ratio = max(value_list) / min(value_list)
                if ratio >= 2:
                    novelty += 0.5
            else:
                novelty += 1
        return -novelty

    @property
    def n_states(self):
        return len(self.states)



import numpy as np
from pymoo.model.crossover import Crossover
import random as rm

from sdc_scissor.testing_api.test_generators.ambiegen.Utils.solution import Solution


class TestCaseCrossover(Crossover):
    """
    Module to perform the crossover
    """

    def __init__(self, cross_rate):
        super().__init__(2, 2)
        self.cross_rate = cross_rate

    def _do(self, problem, X, **kwargs):
        # The input of has the following shape (n_parents, n_matings, n_var)
        _, n_matings, n_var = X.shape

        # The output owith the shape (n_offsprings, n_matings, n_var)
        # Because there the number of parents and offsprings are equal it keeps the shape of X
        Y = np.full_like(X, None, dtype=np.object)

        # for each mating provided

        for k in range(n_matings):

            r = np.random.random()

            s_a, s_b = X[0, k, 0], X[1, k, 0]

            # s_a.get_points()
            # s_a.remove_invalid_cases()
            # s_b.get_points()
            # s_b.remove_invalid_cases()

            if r < self.cross_rate:
                tc_a = s_a.states
                tc_b = s_b.states

                if len(tc_a) < len(tc_b):
                    crossover_point = rm.randint(1, len(tc_a) - 1)
                elif len(tc_b) < len(tc_a):
                    crossover_point = rm.randint(1, len(tc_b) - 1)
                else:
                    crossover_point = rm.randint(1, len(tc_a) - 1)

                if s_a.n_states > 2 and s_b.n_states > 2:

                    offa = {}
                    offb = {}

                    # one point crossover
                    for i in range(0, crossover_point):
                        offa["st" + str(i)] = tc_a["st" + str(i)]
                        offb["st" + str(i)] = tc_b["st" + str(i)]
                    for m in range(crossover_point, len(tc_b)):
                        offa["st" + str(m)] = tc_b["st" + str(m)]
                    for n in range(crossover_point, len(tc_a)):
                        offb["st" + str(n)] = tc_a["st" + str(n)]

                    off_a = Solution()
                    off_b = Solution()

                    off_a.states = offa
                    off_b.states = offb

                else:
                    print("Not enough states!")

                off_a.novelty = off_a.calc_novelty(tc_a, off_a.states)

                off_b.novelty = off_b.calc_novelty(tc_b, off_b.states)

                Y[0, k, 0], Y[1, k, 0] = off_a, off_b

            else:
                Y[0, k, 0], Y[1, k, 0] = s_a, s_b

        return Y



import numpy as np

# TODO: needs improvement in choosing correct number of x, y, theta
def frenet_to_cartesian(x0, y0, theta0, ss, kappas):
    """
    Numerical integration with trapezoidal rule to transform curvature samples
    given in Frenet frame to points in Cartesian frame
    """
    xs = np.zeros(ss.shape[0])
    ys = np.zeros(ss.shape[0])
    thetas = np.zeros(ss.shape[0])
    xs[0] = x0
    ys[0] = y0
    thetas[0] = theta0
    for i in range(thetas.shape[0] - 1):
        thetas[i + 1] = thetas[i] + (kappas[i + 1] + kappas[i]) * (ss[i + 1] - ss[i]) / 2.0
        xs[i + 1] = xs[i] + (ss[i + 1] - ss[i]) * np.cos(thetas[i])
        ys[i + 1] = ys[i] + (ss[i + 1] - ss[i]) * np.sin(thetas[i])
    return (xs, ys)



import logging

from sdc_scissor.obstacle_api.bump import Bump
from sdc_scissor.obstacle_api.beamng_bump import BeamngBump
from sdc_scissor.obstacle_api.delineator import Delineator
from sdc_scissor.obstacle_api.beamng_delineator import BeamngDelineator
from sdc_scissor.obstacle_api.obstacle_factory import ObstacleFactory


class BeamngObstacleFactory(ObstacleFactory):
    def create_bump(self) -> Bump:
        return BeamngBump()

    def create_delineator(self) -> Delineator:
        return BeamngDelineator()


if __name__ == "__main__":
    logging.info("beamng_obstacle_factory.py")



import numpy as np

from self_driving.bbox import RoadBoundingBox

# from code_pipeline.tests_generation import RoadTest
# from code_pipeline.tests_generation import RoadTestFactory
from sdc_scissor.feature_extraction_api.road_geometry_calculator import RoadGeometryCalculator


def find_circle(p1, p2, p3):
    """
    Returns the center and radius of the circle passing the given 3 points.
    In case the 3 points form a line, returns (None, infinity).
    """
    temp = p2[0] * p2[0] + p2[1] * p2[1]
    bc = (p1[0] * p1[0] + p1[1] * p1[1] - temp) / 2
    cd = (temp - p3[0] * p3[0] - p3[1] * p3[1]) / 2
    det = (p1[0] - p2[0]) * (p2[1] - p3[1]) - (p2[0] - p3[0]) * (p1[1] - p2[1])

    if abs(det) < 1.0e-6:
        return np.inf

    # Center of circle
    cx = (bc * (p2[1] - p3[1]) - cd * (p1[1] - p2[1])) / det
    cy = ((p1[0] - p2[0]) * cd - (p2[0] - p3[0]) * bc) / det

    radius = np.sqrt((cx - p1[0]) ** 2 + (cy - p1[1]) ** 2)
    return radius


def min_radius(x, w=5):
    mr = np.inf
    nodes = x
    for i in range(len(nodes) - w):
        p1 = nodes[i]
        p2 = nodes[i + int((w - 1) / 2)]
        p3 = nodes[i + (w - 1)]
        radius = find_circle(p1, p2, p3)
        if radius < mr:
            mr = radius
    if mr == np.inf:
        mr = 0

    return mr * 3.280839895  # , mincurv


class TestValidator:
    def __init__(self, map_size, min_road_length=20):
        self.road_geometry_calculator = RoadGeometryCalculator()
        self.map_size = map_size
        self.box = (0, 0, map_size, map_size)
        self.road_bbox = RoadBoundingBox(self.box)
        self.min_road_length = min_road_length
        # Not sure how to set this value... This might require to compute some sort of density: not points that are too
        # close to each others
        self.max_points = 500

    @staticmethod
    def is_enough_road_points(the_test):
        return len(the_test.road_points) > 1

    def is_too_many_points(self, the_test):
        return len(the_test.road_points) > self.max_points

    @staticmethod
    def is_not_self_intersecting(the_test):
        road_polygon = the_test.get_road_polygon()
        return road_polygon.is_valid()

    @staticmethod
    def is_too_sharp(the_test, TSHD_RADIUS=47):
        return TSHD_RADIUS > min_radius(the_test.interpolated_road_points) > 0.0

    def is_inside_map(self, the_test):
        """
        Take the extreme points and ensure that their distance is smaller than the map side
        """
        xs = [t[0] for t in the_test.interpolated_road_points]
        ys = [t[1] for t in the_test.interpolated_road_points]

        min_x, max_x = min(xs), max(xs)
        min_y, max_y = min(ys), max(ys)

        return (
            min_x > 0
            or min_x > self.map_size
            and max_x > 0
            or max_x > self.map_size
            and min_y > 0
            or min_y > self.map_size
            and max_y > 0
            or max_y > self.map_size
        )

    @staticmethod
    def is_right_type(the_test):
        """
        The type of the_test must be RoadTest
        """
        # check = isinstance(the_test, RoadTestFactory.RoadTest)
        # return check

    @staticmethod
    def is_valid_polygon(the_test):
        road_polygon = the_test.get_road_polygon()
        check = road_polygon.is_valid()
        return check

    def intersects_boundary(self, the_test):
        road_polygon = the_test.get_road_polygon()
        check = self.road_bbox.intersects_boundary(road_polygon.polygon)
        return check

    def is_minimum_length(self, the_test):
        # This is approximated because at this point the_test is not yet interpolated
        return the_test.get_road_length() > self.min_road_length

    @staticmethod
    def is_min_road_segment_not_long_enough_according_risk_factor(the_test):
        rf = the_test.risk_factor
        msl = the_test.min_segment_length
        required_min_segment_length = None
        if rf == 1:
            required_min_segment_length = 50
        elif rf == 1.5:
            required_min_segment_length = 30
        elif rf == 2:
            required_min_segment_length = 20
        else:
            raise Exception("Min seg length for RF {} not defined".format(rf))

        return msl >= required_min_segment_length

    def is_road_not_long_enough_according_min_segment(self, the_test):
        required_proportion = 0.4  # TODO: avoid magic number
        road_length = self.road_geometry_calculator.get_road_length(the_test.road_points)
        msl = the_test.min_segment_length
        return road_length * required_proportion >= msl

    def validate_test(self, the_test):

        is_valid = True
        validation_msg = ""

        # if not self.is_right_type(the_test):
        #     is_valid = False
        #     validation_msg = "Wrong type"
        #     return is_valid, validation_msg

        if not self.is_enough_road_points(the_test):
            is_valid = False
            validation_msg = "Not enough road points."
            return is_valid, validation_msg

        if self.is_too_many_points(the_test):
            is_valid = False
            validation_msg = "The road definition contains too many points"
            return is_valid, validation_msg

        if not self.is_inside_map(the_test):
            is_valid = False
            validation_msg = "Not entirely inside the map boundaries"
            return is_valid, validation_msg

        # if self.intersects_boundary(the_test):
        #     is_valid = False
        #     validation_msg = "Not entirely inside the map boundaries"
        #     return is_valid, validation_msg
        #
        # if not self.is_valid_polygon(the_test):
        #     is_valid = False
        #     validation_msg = "The road is self-intersecting"
        #     return is_valid, validation_msg
        #
        # if not self.is_minimum_length(the_test):
        #     is_valid = False
        #     validation_msg = "The road is not long enough."
        #     return is_valid, validation_msg
        #
        # if self.is_too_sharp(the_test):
        #     is_valid = False
        #     validation_msg = "The road is too sharp"
        #     return is_valid, validation_msg

        # if self.is_min_road_segment_not_long_enough_according_risk_factor(the_test):
        #     is_valid = False
        #     validation_msg = "Min road segment is too short for the risk factor"
        #     return is_valid, validation_msg

        # if self.is_road_not_long_enough_according_min_segment(the_test):
        #     is_valid = False
        #     validation_msg = "Road is too short according the min segment"
        #     return is_valid, validation_msg

        return is_valid, validation_msg



import logging
import time

from scipy.spatial import distance
from shapely.geometry import box

from sdc_scissor.testing_api.test import Test
from sdc_scissor.testing_api.road_model import RoadModel
from sdc_scissor.simulator_api.abstract_simulator import AbstractSimulator


def _get_t_previous_data(data, time_delta) -> tuple:
    """
    Return data of t-delta
    """
    logging.info("_get_t_previous_data")
    t1, _, _, _ = data[-1]
    reversed_data_iterator = reversed(data)
    for t, x_pos, y_pos, z_pos in reversed_data_iterator:
        if t1 - t > time_delta:
            return t, x_pos, y_pos, z_pos
    logging.info("Not enough data. An exception will be thrown.")
    raise Exception("Not enough data")


class TestMonitor:
    """
    The test monitor checks the execution states of the test and logs them.
    """

    def __init__(self, simulator: AbstractSimulator, test: Test, oob: float, road_model: RoadModel):
        """
        The test monitor retrieves data from the simulator and tracks the trajectory of the car.

        :param simulator: The simulator object to get the data from.
        :param test: The test object.
        :param oob: Parameter to define how much percentage the car is allowed to be off the lane.
        """
        self.simulator = simulator
        self.test = test
        self.is_test_finished = False
        self.is_car_out_of_lane = False
        self.road = None
        self.start_time = None
        self.end_time = None
        self.road_model = road_model
        self.oob = oob
        self.has_test_failed = None
        self.current_test_outcome = "UNDEFINED"

    def check(self, interrupt_on_failure):
        """
        Checks the current state of the vehicle and test execution.

        :param interrupt_on_failure: Boolean flag if the test should be interrupted when the car drives off the lane.
        """
        logging.info("check")
        self.simulator.update_car()
        x_pos, y_pos, z_pos = self.simulator.get_car_position()
        current_time = time.time() - self.start_time
        logging.info("time: {}\tx: {}\ty: {}\tz: {}".format(current_time, x_pos, y_pos, z_pos))
        self.test.simulation_data.append((current_time, x_pos, y_pos, z_pos))

        if (
            self.__is_car_at_end_of_road(x_pos, y_pos)
            or (self.__is_car_out_of_lane(x_pos, y_pos) and interrupt_on_failure)
            or not self.is_car_moving()
        ):
            logging.info("TEST IS FINISHED!")
            self.is_test_finished = True
            self.end_time = time.time()
            self.test.test_outcome = self.current_test_outcome
            self.test.test_duration = self.end_time - self.start_time

    def is_car_moving(self) -> bool:
        """
        Checks if the car is currently moving.
        """
        time_delta = 10
        decision_distance = 1

        current_time, current_x_pos, current_y_pos, _ = self.test.simulation_data[-1]
        start_time, _, _, _ = self.test.simulation_data[0]

        # We need at least of `time_delta` seconds of simulation data.
        if current_time - start_time < time_delta:
            return True

        _, last_x_pos, last_y_pos, _ = _get_t_previous_data(self.test.simulation_data, time_delta)

        is_car_moving = not self.__are_points_close(
            (current_x_pos, current_y_pos), (last_x_pos, last_y_pos), decision_distance
        )

        if not is_car_moving:
            logging.warning("CAR IS NOT MOVING!")
            self.has_test_failed = True
            self.current_test_outcome = "FAIL"
        return is_car_moving

    def start_timer(self):
        """
        Start the timer for the test execution.
        """
        logging.info("start_timer")
        self.start_time = time.time()

    def stop_timer(self):
        """
        Stop the timer for the test execution.
        """
        logging.info("stop_timer")
        self.end_time = time.time()

    def dump_data(self):
        """ """
        logging.info("dump_data")
        # TODO

    def __is_car_out_of_lane(self, x_pos: float, y_pos: float) -> bool:
        """

        :param x_pos:
        :param y_pos:
        :return:
        """
        logging.info("__is_car_out_of_lane")

        # Shapely box geometry
        car_model = box(x_pos - 1, y_pos - 1, x_pos + 1, y_pos + 1)

        diff = car_model.difference(self.road_model.right_lane)
        actual_oob = diff.area / car_model.area
        is_car_model_inside_road_model: bool = actual_oob < self.oob

        if not is_car_model_inside_road_model:
            logging.warning("CAR IS OFF THE LANE!")
            self.has_test_failed = True
            self.current_test_outcome = "FAIL"
        return not is_car_model_inside_road_model

    def __is_car_at_end_of_road(self, x_pos: float, y_pos: float) -> bool:
        """

        :param x_pos:
        :param y_pos:
        :return:
        """
        logging.info("__car_at_end_of_road")
        road_end_point = self.test.interpolated_road_points[-1]
        x_end, y_end = road_end_point[0], road_end_point[1]
        is_car_at_the_end_of_the_road: bool = self.__are_points_close((x_pos, y_pos), (x_end, y_end), 7)
        if is_car_at_the_end_of_the_road:
            logging.warning("CAR IS AT THE END OF THE ROAD!")
            if not self.has_test_failed:
                self.current_test_outcome = "PASS"
        return is_car_at_the_end_of_the_road

    @staticmethod
    def __are_points_close(a: tuple[float, float], b: tuple[float, float], threshold: float):
        """

        :param a:
        :param b:
        :param threshold:
        :return:
        """
        dist = distance.euclidean(a, b)
        return dist < threshold


if __name__ == "__main__":
    logging.info("test_monitor.py")



import json
import logging
from pathlib import Path

from sdc_scissor.testing_api.test import Test
from sdc_scissor.testing_api.test_generators.ambiegen.ambiegen_generator import CustomAmbieGenGenerator
from sdc_scissor.testing_api.test_generators.frenetic.src.generators.random_frenet_generator import (
    CustomFrenetGenerator,
)
from sdc_scissor.testing_api.test_generators.frenetic_v.src.generators.random_frenet_generator import (
    CustomFrenetVGenerator,
)


def _id_generator():
    cnt: int = 0
    while True:
        yield cnt
        cnt += 1


class TestGenerator:
    def __init__(self, count: int, destination: Path, tool: str):
        """
        This class is used to generate tests for a virtual environment.
        """

        self.count: int = count
        self.__id_generator = _id_generator()
        self.__nr_prefix_digits: int = 5
        self.destination: Path = destination
        self.generated_tests: list[Test] = []
        self.tool: str = tool
        kwargs: dict = {"map_size": 200, "time_budget": 100, "count": count}
        # Types of test generator
        if self.tool.lower() == "frenetic":
            self.random_generator = CustomFrenetGenerator(**kwargs)
        elif self.tool.lower() == "freneticv":
            self.random_generator = CustomFrenetVGenerator(**kwargs)
        elif self.tool.lower() == "ambiegen":
            self.random_generator = CustomAmbieGenGenerator()
        else:
            raise Exception(" Invalid tool name. Supported tools [frenetic, freneticv, ambiegen]")

    def generate(self):
        """
        Generate tests according to the parameters set while instantiating this object.
        """
        logging.debug("* generate")
        generated_tests_as_list_of_road_points = self.random_generator.start()
        generated_tests_as_list_of_road_points = self.__extract_valid_roads(generated_tests_as_list_of_road_points)

        generated_tests_as_list_of_road_points = self.__add_sine_bumps(generated_tests_as_list_of_road_points)

        for road_points in generated_tests_as_list_of_road_points:
            test = Test(test_id=next(self.__id_generator), road_points=road_points, test_outcome="NOT_EXECUTED")
            self.generated_tests.append(test)
        logging.info("** {} tests generated".format(len(generated_tests_as_list_of_road_points)))
        logging.info("** test generator has {} tests".format(len(self.generated_tests)))

    def __add_sine_bumps(self, generated_tests_as_list_of_road_points):
        for road_as_points in generated_tests_as_list_of_road_points:
            pass
        return generated_tests_as_list_of_road_points

    def __extract_valid_roads(self, road_points):
        """
        Check if the road points are actual valid roads without intersections or too sharp turns.

        :param road_points: List of road points specifying potential roads
        :return: List of road points that define proper valid roads
        """
        # TODO
        return road_points

    def save_tests(self):
        """
        Save the tests as json files in a separate directory.
        """
        logging.info("* save_tests")

        file_post_fix: str = "_test.json"
        for i, test in enumerate(self.generated_tests):
            filename = self.__get_next_filename(test, file_post_fix)
            full_path: Path = self.destination / filename
            with open(full_path, "w") as fp:
                test_dict = vars(test)
                json.dump(test_dict, fp, indent=2)

    def __get_next_filename(self, test: Test, file_post_fix):
        """

        :param test:
        :param file_post_fix:
        :return:
        """
        test_id: str = str(test.test_id)
        nr_zeros: int = self.__nr_prefix_digits - len(test_id)
        file_pre_fix: str = ""
        for i in range(nr_zeros):
            file_pre_fix += "0"
        file_pre_fix += test_id
        file_name: str = file_pre_fix + file_post_fix
        return file_name


if __name__ == "__main__":
    logging.info("* test_generator.py")



import logging

import pandas as pd

from pathlib import Path


class CSVLoader:
    @staticmethod
    def load_dataframe_from_csv(data_path: Path):
        """

        :param data_path:
        :return:
        """
        dd = pd.read_csv(data_path)
        return dd


if __name__ == "__main__":
    logging.info("csv_loader.py")



import logging

from beamngpy import ProceduralBump

from sdc_scissor.obstacle_api.bump import Bump


class BeamngBump(Bump):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self):
        return ProceduralBump(
            name="pybump",
            pos=(self.x_pos, self.y_pos, self.z_pos),
            rot=self.rot,
            rot_quat=self.rot_quat,
            width=self.width,
            length=self.length,
            height=self.height,
            upper_length=self.upper_length,
            upper_width=self.upper_width,
        )


if __name__ == "__main__":
    logging.info("beamng_bump.py")



from abc import ABC, abstractmethod
import logging as log
import numpy as np
import pandas as pd
import itertools as it
import os
from pathlib import Path

from time import sleep
from datetime import datetime

# from code_pipeline.tests_generation import RoadTestFactory


class BaseGenerator(ABC):
    def __init__(self, time_budget=None, executor=None, map_size=None, strict_father=False):
        self.time_budget = time_budget
        self.executor = executor
        self.map_size = map_size
        self.df = pd.DataFrame()
        creation_date = datetime.now().strftime("%Y%m%d-%H%M%S")
        self.file_name = f"experiments/{creation_date}-{self.__class__.__name__}-results.csv"
        self.columns_number = 0
        log.info(f"ERATO experiment output is stored in {self.file_name}")

        # Adding mutants for future mutation only if its min_oob_distance is better than its parent's min_oob_distance
        # min_oob_distance < parent_min_oob_distance
        self.strict_father = strict_father

    @abstractmethod
    def start(self):
        pass

    def store_dataframe(self):
        log.info("Storing the all the experiment results in a csv.")
        # Storing the results as csv in experiments folders
        with open(self.file_name, "w") as outfile:
            self.df.to_csv(outfile)

    # def execute_test(self, road_points, method='random', extra_info={}, parent_info={}):
    #     # Some more debugging
    #     log.info("Generated test using: %s", road_points)
    #     the_test = RoadTestFactory.create_road_test(road_points)
    #
    #     # Try to execute the test
    #     test_outcome, description, execution_data = self.executor.execute_test(the_test)
    #
    #     # Print the result from the test and continue
    #     log.info("test_outcome %s", test_outcome)
    #     log.info("description %s", description)
    #     info = {'outcome': test_outcome, 'description': description, 'road': road_points, 'method': method,
    #             'visited': False, 'ancestors': [], 'generation': 0}
    #
    #     # Adding extra info to the dataframe
    #     for k, v in extra_info.items():
    #         info[k] = v
    #
    #     min_oob_distance = None
    #     # Storing the data in a dataframe for next phase
    #     if execution_data:
    #         # base metrics
    #         metrics = ['steering', 'steering_input', 'brake', 'brake_input', 'throttle', 'throttle_input',
    #         'wheelspeed', 'vel_kmh', 'oob_counter', 'oob_distance']
    #         functions = [('max', np.max), ('min', np.min), ('mean', np.mean), ('avg', np.average)]
    #         for metric, (name, func) in it.product(metrics, functions):
    #             metric_data = [y for y in map(lambda x: getattr(x, metric), execution_data) if y is not None]
    #             if metric_data:
    #                 info['{:s}_{:s}'.format(name, metric)] = func(metric_data)
    #         info['max_oob_percentage'] = execution_data[-1].max_oob_percentage
    #
    #         # complex metrics
    #         accum_neg_oob = self.accumulated_negative_oob(execution_data)
    #         info['accum_neg_oob'] = accum_neg_oob
    #
    #         # parent info
    #         for k, v in parent_info.items():
    #             info[k] = v
    #
    #         # avoid visiting mutants that perform worst than its parents
    #         if self.strict_father and parent_info and info['min_oob_distance'] > info['parent_min_oob_distance']:
    #             info['visited'] = True
    #             log.info('Weaker mutant: Disabling current test for future mutations.')
    #
    #         # Retrieving file name
    #         last_file = sorted(Path('simulations/beamng_executor').iterdir(), key=os.path.getmtime)[-1]
    #         info['simulation_file'] = last_file.name
    #
    #         min_oob_distance = info['min_oob_distance']
    #
    #         # Logging some info for debugging
    #         log.info('Min oob_distance: {:0.3f}'.format(info['min_oob_distance']))
    #         log.info('Accumulated negative oob_distance: {:0.3f}'.format(accum_neg_oob))
    #
    #     self.df = self.df.append(info, ignore_index=True)
    #
    #     if self.executor.road_visualizer:
    #         sleep(5)
    #     return info['outcome'], min_oob_distance

    @staticmethod
    def accumulated_negative_oob(execution_data):
        """
        Note: Normalizing oob_distance to be negative.
        Default interval: [-2, 2] --> Normalized interval: [-4, 0]
        Args:
            execution_data: execution data from the simulator
        Returns:
            Accumulated oob_distance when the center the mass already crossed one of the lanes (oob_distance < 0).
        """
        return sum(
            map(
                lambda k: (execution_data[k].oob_distance - 2) * (execution_data[k].timer - execution_data[k - 1].timer)
                if execution_data[k].oob_distance < 0
                else 0,
                range(1, len(execution_data)),
            )
        )



from sdc_scissor.feature_extraction_api.segmentation_strategy import SegmentationStrategy


class EquiDistanceStrategy(SegmentationStrategy):
    """
    Define segments with equal lengts. The last segment might be different to
    the others since there might be a remainder while dividing the road into
    predefined number of segments.
    """

    def __init__(self, number_of_segments):
        """
        Initialize the strategy by providing the number of segments that should be generated

        :param number_of_segments: The number of segments with equal distances that should be generated
        """
        super().__init__()
        self.__number_of_segments = number_of_segments

    def extract_segments(self, road_points):
        """
        Extract the segments of the road specified by the road points

        :param road_points: Road points defining the road
        :return: List of indexes representing the start and end point of segments
        """
        segments = []

        # only two road points for a segment
        max_number_of_possible_segments = len(road_points) - 1

        number_of_road_points = len(road_points)

        if number_of_road_points < self.__number_of_segments:
            raise Exception("Not enough road points.")

        # TODO: Verify if this calculation is correct. I am not sure!
        road_points_per_segment = 1 + round(max_number_of_possible_segments / self.__number_of_segments)

        # calculate for each segments its start/end road point indeces
        end = 0
        for i in range(self.__number_of_segments):
            start = end

            # check if last segment because last segment might have more/less points
            if i == self.__number_of_segments - 1:
                end = len(road_points) - 1
            else:
                end = start + road_points_per_segment - 1

            current_segment = (start, end)
            segments.append(current_segment)

        return segments


if __name__ == "__main__":
    import unittest

    class EquiDistanceSegmentationTest(unittest.TestCase):
        def setUp(self):
            pass

        def test_n5_6_straight_points(self):
            nr_of_segments = 5
            strategy = EquiDistanceStrategy(nr_of_segments)

            road_points = [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1)]

            segments = strategy.extract_segments(road_points)

            self.assertEqual(len(segments), 5, "There must be {} segments.".format(nr_of_segments))

            expected_segments = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
            self.assertEqual(segments, expected_segments, "Segments are not correctly calculated.")

        def test_too_few_road_points_should_raise_exception(self):
            nr_of_segments = 4
            strategy = EquiDistanceStrategy(nr_of_segments)

            # 6 road points
            road_points = [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1)]

            self.assertRaises(Exception, strategy.extract_segments(road_points))

        def test_last_segment_has_more_road_points_than_others(self):
            nr_of_segments = 4
            strategy = EquiDistanceStrategy(nr_of_segments)

            road_points = [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1)]

            segments = strategy.extract_segments(road_points)

            expected_segments = [(0, 1), (1, 2), (2, 3), (3, 5)]
            self.assertEqual(segments, expected_segments, "Segments are not correctly calculated.")

        def test_last_segment_has_less_road_points_than_others(self):
            nr_of_segments = 3
            strategy = EquiDistanceStrategy(nr_of_segments)

            road_points = [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1)]

            segments = strategy.extract_segments(road_points)

            expected_segments = [(0, 2), (2, 4), (4, 5)]
            self.assertEqual(segments, expected_segments, "Segments are not correctly calculated.")

    unittest.main()



import logging

from sdc_scissor.obstacle_api.bump import Bump
from sdc_scissor.obstacle_api.carla_bump import CarlaBump
from sdc_scissor.obstacle_api.delineator import Delineator
from sdc_scissor.obstacle_api.carla_delineator import CarlaDelineator
from sdc_scissor.obstacle_api.obstacle_factory import ObstacleFactory


class CarlaObstacleFactory(ObstacleFactory):
    def create_bump(self) -> Bump:
        return CarlaBump()

    def create_delineator(self) -> Delineator:
        return CarlaDelineator()


if __name__ == "__main__":
    logging.info("carla_obstacle_factory.py")



import logging
import abc


class Bump(abc.ABC):
    def __init__(
        self,
        x_pos=None,
        y_pos=None,
        z_pos=None,
        width=6,
        length=2,
        height=0.2,
        upper_length=1,
        upper_width=2,
        rot=None,
        rot_quat=(0, 0, 0, 1),
    ):
        super().__init__()
        self.x_pos = x_pos
        self.y_pos = y_pos
        self.z_pos = z_pos
        self.width = width
        self.length = length
        self.height = height
        self.upper_length = upper_length
        self.upper_width = upper_width
        self.rot = rot
        self.rot_quat = rot_quat

    @abc.abstractmethod
    def get(self):
        pass


if __name__ == "__main__":
    logging.info("bump.py")



import csv
import os

selected_rows = []
remove = False

final_csv_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "BeamNG_RF_1_5_selected.csv")
with open(final_csv_dir, "r") as file:
    reader = csv.reader(file)
    for row in reader:
        if row[19] == "safe":
            selected_rows.append(row)
        else:
            if not remove:
                selected_rows.append(row)

            remove = not remove

print(len(selected_rows))

final_csv_file = open(final_csv_dir, "wb")
final_csv_writer = csv.writer(final_csv_file)

fields = [
    "direct_distance",
    "road_distance",
    "num_l_turns",
    "num_r_turns",
    "num_straights",
    "median_angle",
    "total_angle",
    "mean_angle",
    "std_angle",
    "max_angle",
    "min_angle",
    "median_pivot_off",
    "mean_pivot_off",
    "std_pivot_off",
    "max_pivot_off",
    "min_pivot_off",
    "start_time",
    "end_time",
    "duration_seconds",
    "safety",
]
final_csv_writer.writerow(fields)

for row in selected_rows:
    print(row)
    final_csv_writer.writerow(row)



import logging
import abc


class Delineator(abc.ABC):
    def __init__(self, x_pos=None, y_pos=None, z_pos=None, radius=0.2, height=1.0, rot=None, rot_quat=(0, 0, 0, 1)):
        super().__init__()
        self.x_pos = x_pos
        self.y_pos = y_pos
        self.z_pos = z_pos
        self.radius = radius
        self.height = height
        self.rot = rot
        self.rot_quat = rot_quat

    @abc.abstractmethod
    def get(self):
        pass


if __name__ == "__main__":
    logging.info("delineator.py")



import os

# dirs
codes_directory = os.path.dirname(os.path.abspath(__file__))
root_directory = os.path.join(codes_directory, "..", "..")
data_directory = os.path.join(root_directory, "data")

# benchmarks, configs, and number of execution repeats used in this experiment
benchmarks = ["BeamNG_RF_1_Complete", "BeamNG_RF_1_5_selected", "DriverAI_Complete"]
configurations = ["default-GA", "10_feature_GA", "6_feature_GA"]
number_of_runs = 30

# Open the output csv
out_csv_address = os.path.join(data_directory, "results.csv")
out_csv = open(out_csv_address, "w")
out_csv.write("benchmark,config,execution_id,best_fitness_value,APFD,elapsed_time,avg_rand_APFD,std_rand_APFD")
out_csv.write("\n")
out_csv.close()

out_csv = open(out_csv_address, "a")
for benchmark in benchmarks:
    for config in configurations:
        for run in range(1, number_of_runs + 1):
            current_data_directory = os.path.join(data_directory, benchmark, config, str(run))
            current_results_csv = os.path.join(current_data_directory, "results.csv")

            f = open(current_results_csv)
            lines = f.readlines()
            # pop the title line
            lines.pop(0)

            for interesting_line in lines:
                prepared_line_to_save = benchmark + "," + interesting_line
                out_csv.write(prepared_line_to_save)

out_csv.close()



import logging
import json

import numpy as np

from scipy.interpolate import splprep, splev
from pathlib import Path


class Test:
    def __init__(self, test_id, road_points: list[list], test_outcome, test_duration=None):
        """
        Class representing a test case.

        :param test_id: Unique test identifier
        :param road_points: Road points defining the test
        :param test_outcome: Outcome of the test execution
        :param test_duration: The duration of the test execution
        """
        self.test_id = test_id
        self.test_outcome = test_outcome
        self.predicted_test_outcome = None
        self.test_duration = test_duration
        self.road_points = road_points
        self.interpolated_road_points = self.__interpolate(road_points)
        self.simulation_data = []

    def save_as_json(self, file_path: Path):
        """

        :param file_path: File path to save the test as a json file
        """
        logging.info("save_as_json")
        with open(file_path, "w") as fp:
            test_dict = vars(self)
            json.dump(test_dict, fp, indent=2)

    @staticmethod
    def __interpolate(road_points: list[list]):
        """

        :param road_points:
        :return:
        """
        logging.info("* __interpolate")
        road_matrix = np.array(road_points)
        x = road_matrix[:, 0]
        y = road_matrix[:, 1]
        num_nodes = 50

        if len(x) == 2:
            # With two points the only option is a straight segment
            k = 1
        elif len(y) == 3:
            # With three points we use an arc, using linear interpolation will result in invalid road tests
            k = 2
        else:
            # Otheriwse, use cubic splines
            k = 3

        pos_tck, *_pos_u = splprep([x, y], s=0, k=k)
        step_size = 1 / num_nodes
        unew = np.arange(0, 1 + step_size, step_size)
        x_new, y_new = splev(unew, pos_tck)
        new_road_points = np.column_stack((x_new, y_new)).tolist()

        return new_road_points


if __name__ == "__main__":
    logging.info("* test.py")



from shapely.geometry import LineString


class RoadModel:
    def __init__(self, road_points: list[list]):
        self.coordinates = [(x[0], x[1]) for x in road_points]
        self.center_line: LineString = LineString(coordinates=self.coordinates)
        self.right_lane = self.center_line.buffer(distance=-5, single_sided=True)
        self.ideal_trajectory: LineString = self.center_line.parallel_offset(distance=2.5)



import logging
import joblib

import pandas as pd
import numpy as np

from pathlib import Path
from sklearn import preprocessing
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.model_selection import KFold, cross_validate, train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import LinearSVC
from sklearn.metrics import accuracy_score, recall_score, f1_score, precision_score


class ModelEvaluator:
    def __init__(
        self,
        data_frame: pd.DataFrame,
        label: str,
        features=(
            "full_road_diversity",
            "mean_road_diversity",
            "direct_distance",
            "max_angle",
            "max_pivot_off",
            "mean_angle",
            "mean_pivot_off",
            "median_angle",
            "median_pivot_off",
            "min_angle",
            "min_pivot_off",
            "num_l_turns",
            "num_r_turns",
            "num_straights",
            "road_distance",
            "std_angle",
            "std_pivot_off",
            "total_angle",
        ),
    ):
        """

        :param data_frame:
        :param label:
        :param features:
        """
        self.data_frame = data_frame
        self.label = label
        self.road_features = list(features)

        self.__classifiers = {}

    def evaluate(self):
        """

        :return:
        """
        logging.info("evaluate")

        dd = self.data_frame.sample(frac=1).reset_index(drop=True)

        attributes_to_use = self.road_features.copy()
        attributes_to_use.append(self.label)
        dd = dd[attributes_to_use]
        dd = dd.dropna()

        N = dd.shape[0]
        N_train = int(N * 0.8)
        N_test = N - N_train

        X = dd[attributes_to_use[:-1]].to_numpy()
        X_train = X[0:N_train, :]
        X_test = X[N_train:, :]

        X = preprocessing.normalize(X)
        X = preprocessing.scale(X)

        y = dd[attributes_to_use[-1]].to_numpy()
        y[y == "FAIL"] = 1
        y[y == "PASS"] = 0
        y = np.array(y, dtype="int32")

        y_train = y[:N_train]

        X_train_pass = X_train[y_train == 0, :]
        X_train_fail = X_train[y_train == 1, :]
        y_train_pass = y_train[y_train == 0]
        y_train_fail = y_train[y_train == 1]
        y_test = y[N_train:]

        n_fail = len(y_train_fail)
        n_pass = len(y_train_pass)
        if n_fail < n_pass:
            y_train = np.concatenate((y_train_fail, y_train_pass[:n_fail]))
            X_train = np.concatenate((X_train_fail, X_train_pass[:n_fail, :]), axis=0)
        else:
            y_train = np.concatenate((y_train_pass, y_train_fail[:n_pass]))
            X_train = np.concatenate((X_train_pass, X_train_fail[:n_pass, :]), axis=0)

        self.__classifiers["random_forest"] = RandomForestClassifier()
        self.__classifiers["gradient_boosting"] = GradientBoostingClassifier()
        self.__classifiers["SVM"] = LinearSVC(max_iter=10000)
        self.__classifiers["gaussian_naive_bayes"] = GaussianNB()
        self.__classifiers["logistic_regression"] = LogisticRegression(max_iter=10000)
        self.__classifiers["decision_tree"] = DecisionTreeClassifier()

        for model_name, model in self.__classifiers.items():
            model.fit(X_train, y_train)
            y_pred = model.predict(X_test)

            acc = accuracy_score(y_test, y_pred)
            prec = precision_score(y_test, y_pred)
            rec = recall_score(y_test, y_pred)
            f1 = f1_score(y_test, y_pred)

            print(
                "MODEL: {:<25} ACCURACY: {:<20} RECALL: {:<20} PRECISION: {:<20} F1: {}".format(
                    model_name, acc, rec, prec, f1
                )
            )

    def save_models(self, out_dir: Path):
        """

        :param out_dir:
        """
        logging.info("save_models")

        for model_name, model in self.__classifiers.items():
            filename = model_name + ".joblib"
            file_path = out_dir / filename
            logging.info("save model: {}".format(model_name))
            joblib.dump(model, file_path)



import logging
import sys
import click
import yaml

from pathlib import Path

from sdc_scissor.simulator_api.simulator_factory import SimulatorFactory
from sdc_scissor.testing_api.test_runner import TestRunner
from sdc_scissor.testing_api.test_generator import TestGenerator
from sdc_scissor.testing_api.test_loader import TestLoader
from sdc_scissor.feature_extraction_api.feature_extraction import FeatureExtractor
from sdc_scissor.feature_extraction_api.angle_based_strategy import AngleBasedStrategy
from sdc_scissor.machine_learning_api.csv_loader import CSVLoader
from sdc_scissor.machine_learning_api.model_evaluator import ModelEvaluator
from sdc_scissor.machine_learning_api.cost_effectiveness_evaluator import CostEffectivenessEvaluator
from sdc_scissor.machine_learning_api.predictor import Predictor
from sdc_scissor.obstacle_api.beamng_obstacle_factory import BeamngObstacleFactory

_ROOT_DIR = Path(__file__).parent
_DESTINATION = _ROOT_DIR / "destination"
_TRAINED_MODELS = _ROOT_DIR / "trained_models"


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("-c", "--config", type=click.Path(exists=True), help="Configuration file")
@click.option("--debug/--no-debug", default=False)
def cli(ctx: click.Context, config: Path, debug) -> None:
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if ctx.invoked_subcommand:
        return None
    logging.info("run with configuration file")
    with open(config) as fp:
        config_dict: dict = yaml.safe_load(fp)
        command = config_dict["command"]
        options: dict = config_dict["options"]
        this_module = sys.modules[__name__]
        command = getattr(this_module, command.replace("-", "_"))
        ctx.invoke(command, **options)


@cli.command()
@click.option("-c", "--count", type=int, default=10)
@click.option("-d", "--destination", default=_DESTINATION, type=click.Path())
@click.option("-t", "--tool", default="frenetic", type=click.STRING)
def generate_tests(count: int, destination: Path, tool: str) -> None:
    """
    Generate tests (road specifications) for self-driving cars.
    """
    logging.info("generate_tests")
    destination = Path(destination)
    if not destination.exists():
        destination.mkdir(parents=True)

    test_generator = TestGenerator(count=count, destination=destination, tool=tool)
    test_generator.generate()
    test_generator.save_tests()


@cli.command()
@click.option("-t", "--tests", default=_DESTINATION, type=click.Path(exists=True))
@click.option("-s", "--segmentation", default="angle-based", type=click.STRING)
def extract_features(tests: Path, segmentation: str) -> None:
    """
    Extract road features from given test scenarios.
    """
    logging.info("extract_features")
    tests = Path(tests)
    test_loader = TestLoader(tests)
    if segmentation == "angle-based":
        segmentation = AngleBasedStrategy(angle_threshold=5, decision_distance=10)
    feature_extractor = FeatureExtractor(segmentation_strategy=segmentation)
    road_features_lst = []
    while test_loader.has_next():
        test, _ = test_loader.next()
        road_features = feature_extractor.extract_features(test)
        road_features.safety = test.test_outcome
        road_features_lst.append((test.test_id, road_features, test.test_duration))

    FeatureExtractor.save_to_csv(road_features_lst, tests)


@cli.command()
@click.option("-t", "--tests", default=_DESTINATION, type=click.Path(exists=True))
@click.option("--home", type=click.Path(exists=True))
@click.option("--user", type=click.Path(exists=True))
@click.option("--rf", default=1.5, type=float)
@click.option("--oob", default=0.3, type=float)
@click.option("--max-speed", default=50, type=float)
@click.option("--interrupt/--no-interrupt", default=True, type=click.BOOL)
@click.option("--obstacles/--no-obstacles", default=False, type=click.BOOL)
@click.option("--bump-dist", default=20, type=click.INT)
@click.option("--delineator-dist", default=5, type=click.INT)
@click.option("-fov", "--field-of-view", default=120, type=click.INT)
def label_tests(
    tests: Path, home, user, rf, oob, max_speed, interrupt, obstacles, bump_dist, delineator_dist, field_of_view
) -> None:
    """
    Execute the tests in simulation to label them as safe or unsafe scenarios.
    """
    logging.info("label_tests")
    tests = Path(tests)
    logging.debug("Test directory: {}".format(tests))
    beamng_simulator = SimulatorFactory.get_beamng_simulator(
        home=home, user=user, rf=rf, max_speed=max_speed, fov=field_of_view
    )

    test_loader = TestLoader(tests_dir=tests)

    if obstacles:
        obstacle_factory = BeamngObstacleFactory()
    else:
        obstacle_factory = None

    test_runner = TestRunner(
        simulator=beamng_simulator,
        test_loader=test_loader,
        oob=oob,
        interrupt=interrupt,
        obstacle_factory=obstacle_factory,
        bump_dist=bump_dist,
        delineator_dist=delineator_dist,
    )

    test_runner.run_test_suite()


@cli.command()
@click.option("--csv", default=_DESTINATION / "road_features.csv", type=click.Path(exists=True))
@click.option("--models-dir", default=_TRAINED_MODELS, type=click.Path())
def evaluate_models(csv: Path, models_dir: Path) -> None:
    """
    Evaluate different machine learning models.
    """
    logging.info("evaluate_models")

    if not models_dir.exists():
        models_dir.mkdir()

    dd = CSVLoader.load_dataframe_from_csv(csv)

    model_evaluator = ModelEvaluator(data_frame=dd, label="safety")
    model_evaluator.evaluate()
    model_evaluator.save_models(out_dir=models_dir)


@cli.command()
@click.option(
    "--csv", default=_DESTINATION / "road_features.csv", help="Path to labeled tests", type=click.Path(exists=True)
)
@click.option("--train-ratio", default=0.7, help="Ratio used for training the models", type=click.FLOAT)
def evaluate_cost_effectiveness(csv: Path, train_ratio: float) -> None:
    """
    Evaluate the speed-up SDC-Scissor achieves by only selecting test scenarios that likely fail.
    """
    logging.info("evaluate_cost_effectiveness")
    dd = CSVLoader.load_dataframe_from_csv(csv)

    df = dd.sample(frac=1).reset_index(drop=True)

    cost_effectiveness_evaluator = CostEffectivenessEvaluator(data_frame=df, label="safety", time_attribute="duration")
    cost_effectiveness_evaluator.evaluate()


@cli.command()
@click.option("-t", "--tests", default=_DESTINATION, type=click.Path(exists=True))
@click.option("-c", "--classifier", default=_TRAINED_MODELS / "decision_tree.joblib", type=click.Path(exists=True))
def predict_tests(tests: Path, classifier: Path) -> None:
    """
    Predict the most likely outcome of a test scenario without executing them in simulation.
    """
    test_loader = TestLoader(tests_dir=tests)

    predictor = Predictor(test_loader=test_loader, joblib_classifier=classifier)
    predictor.predict()


if __name__ == "__main__":
    cli()



import abc
import logging


class AbstractSimulator(abc.ABC):
    def __init__(self):
        """ """
        super().__init__()

    @abc.abstractmethod
    def open(self):
        """
        Start the simulator in a separate process and enable interprocess communication by opening a socket.
        """
        pass

    @abc.abstractmethod
    def close(self):
        """
        Quit the simulator process and close the socket.
        """
        pass

    @abc.abstractmethod
    def create_new_instance(self):
        """
        Restart the simulator process.
        """
        pass

    @abc.abstractmethod
    def stop_scenario(self):
        """
        Stop the execution of an ongoing scenario in simulation.
        """
        pass

    @abc.abstractmethod
    def start_scenario(self):
        """
        Start the execution of a scenario in simulation.
        """
        pass

    @abc.abstractmethod
    def load_scenario(self, test, obstacles: list):
        """
        Prepare the simulator for a specific test scenario.

        :param test: An object defining the test scenario, i.e. vehicle setup, road, etc.
        :param obstacles: A list with obstacles to place in the virtual environment.
        """
        pass

    @abc.abstractmethod
    def update_car(self):
        """
        Retrieve the current sensor data of the vehicle in a running test that is executing in simulation within another process.
        """
        pass

    @abc.abstractmethod
    def get_car_position(self):
        """ """
        pass


if __name__ == "__main__":
    logging.info("* abstract_simulator.py")



import logging
import re
import json
import os
from pathlib import Path

from sdc_scissor.testing_api.test import Test


class TestLoader:
    def __init__(self, tests_dir: Path):
        """

        :param tests_dir:
        """
        self.tests_dir: Path = tests_dir

        self.test_paths: list[Path] = []
        self.__set_test_paths(self.test_paths)

    def __set_test_paths(self, tests_paths: list):
        """

        :param tests_paths:
        """
        logging.debug("* _test_loader_gen")
        pattern: str = r".*test.*.json"
        for root, dirs, files in os.walk(self.tests_dir):
            if "unlabeled" in root:
                continue
            for file in files:
                if re.fullmatch(pattern, file):
                    full_path = Path(root) / file
                    tests_paths.append(full_path)

    def has_next(self) -> bool:
        """

        :return:
        """
        return len(self.test_paths) > 0

    def next(self) -> tuple[Test, Path]:
        """

        :return:
        """
        if not self.has_next():
            logging.warning("There are no remaining tests!")
            raise Exception("There are no remaining tests!")

        test_path: Path = self.test_paths.pop()
        test: Test = self.__load_test_from_path(test_path)
        return test, test_path

    @staticmethod
    def __load_test_from_path(test_path: Path):
        """

        :param test_path:
        :return:
        """
        logging.info(str(test_path))
        with open(test_path, "r") as fp:
            test_json: dict = json.load(fp)

        road_points = test_json.get("interpolated_road_points", None)
        test_outcome = test_json.get("test_outcome", None)
        sim_time = test_json.get("simulation_time", None)

        id_pattern = r"(.*test.*)"
        logging.info("test_path: {}".format(str(test_path)))
        match_obj = re.match(pattern=id_pattern, string=str(test_path))
        test_id = match_obj.group(1)
        logging.info("test_id: {}".format(test_id))
        test = Test(test_id=test_id, road_points=road_points, test_outcome=test_outcome, test_duration=sim_time)
        return test


if __name__ == "__main__":
    logging.info("* test_loader.py")



from pymoo.core.crossover import Crossover
import numpy as np
import random


def pmx(parent1, parent2):
    cutPoint = random.randint(0, len(parent1) - 1)
    head = parent1[:cutPoint]
    tail = []
    for test_id in parent2:
        if test_id not in head:
            tail.append(test_id)
    tail = np.array(tail)

    offspring = np.concatenate((head, tail), axis=None)

    return offspring


class PMXCrossover(Crossover):
    def __init__(self, **kwargs):
        super().__init__(2, 1, **kwargs)

    def _do(self, problem, X, **kwargs):
        _, n_matings, n_var = X.shape
        Y = np.full((self.n_offsprings, n_matings, n_var), -1, dtype=int)

        for i in range(n_matings):
            a, b = X[:, i, :]
            Y[0, i, :] = pmx(a, b)

        return Y



# from code_pipeline.tests_generation import RoadTestFactory
from time import sleep
import logging as log

from pymoo.optimize import minimize

from sdc_scissor.testing_api.test_generators.ambiegen.Utils.test_case_problem import TestCaseProblem
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.test_case_mutation import TestCaseMutation
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.test_case_crossover import TestCaseCrossover
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.duplicate_elimination import DuplicateElimination
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.generate_test_case_sampling import GenerateTestCaseSampling
import time
from pymoo.algorithms.nsga2 import NSGA2
import sdc_scissor.testing_api.test_generators.ambiegen.config as cf


class CustomAmbieGenGenerator:
    """
    This test generator creates road points using affine tratsformations to vectors.
    Initially generated test cases are optimized by NSGA2 algorithm with two objectives:
    fault revealing power and diversity. We use a simplified model of a vehicle to
    estimate the fault revealing power (as the maximum deviation from the road center).
    We use 100 generations and 100 population size. In each iteration of the generator
    the Pareto optimal solutions are provided and executed. Then the algorithm is launched again.
    """

    def __init__(self, time_budget=None, executor=None, map_size=200):
        self.map_size = map_size
        self.time_budget = time_budget
        self.executor = executor

    def start(self):
        """
        In this function the algorithm is launched and
        the Pareto optimal solutions are returned
        """

        log.info("Test generation ambiegen.")
        algorithm = NSGA2(
            n_offsprings=50,
            pop_size=cf.ga["population"],
            sampling=GenerateTestCaseSampling(),
            crossover=TestCaseCrossover(cf.ga["cross_rate"]),
            mutation=TestCaseMutation(cf.ga["mut_rate"]),
            eliminate_duplicates=DuplicateElimination(),
        )

        t = int(time.time() * 1000)
        seed = ((t & 0xFF000000) >> 24) + ((t & 0x00FF0000) >> 8) + ((t & 0x0000FF00) << 8) + ((t & 0x000000FF) << 24)

        res = minimize(
            TestCaseProblem(),
            algorithm,
            ("n_gen", cf.ga["n_gen"]),
            seed=seed,
            verbose=False,
            save_history=True,
            eliminate_duplicates=True,
        )

        # print("Best solution found: \nF = %s" % (res.F))
        gen = len(res.history) - 1
        test_cases = []
        i = 0

        while i < len(res.F):
            result = res.history[gen].pop.get("X")[i]

            road_points = result[0].intp_points
            test_cases.append(road_points)
            i += 1
        return test_cases



from .segmentation_strategy import SegmentationStrategy
from .road_geometry_calculator import RoadGeometryCalculator

seg_lengths_dict = {"1": 50, "1.5": 30, "2": 10}


class ParameterizedUniformStrategy(SegmentationStrategy):
    """
    This class segments the road based on parameters like risk factor and
    fullroad length.
    """

    def __init__(self, risk_factor: str, max_seg_length_to_full_road):
        """

        :param risk_factor: Risk factor of the driving AI.
        :param max_seg_length_to_full_road: Relative size of the longest allowed segment compared to the full road length.
        """
        # dependend on the risk factor
        self.__seg_length_in_meters = seg_lengths_dict[risk_factor]
        # dependend on the length of the full road
        self.__max_seg_length_in_meters = None

        self.__max_seg_length_to_full_road = max_seg_length_to_full_road

        self.__road_geometry_calculator = RoadGeometryCalculator()

    # TODO: formalize the clear segmentaion process
    def extract_segments(self, road_points):
        """
        Returns a list of tuples that contain start end end indexes of road
        segments.

        :param road_points: List of coordinates specifying the road.
        :return: List of start and end indexes defining the start and end  road points of segments.
        """
        road_length = self.__road_geometry_calculator.get_road_length(road_points)

        self.__max_seg_length_in_meters = road_length * self.__max_seg_length_to_full_road

        if self.__seg_length_in_meters > self.__max_seg_length_in_meters:
            raise Exception("Road is too short for the required segment length.")

        # iterate over road points until we reach the desired segment length
        end = 0
        start = 0
        current_segment_length = 0
        segment_indexes = []
        for i in range(1, len(road_points)):
            current_elementary_segment = road_points[i - 1 : i + 1]
            current_segment_length += self.__road_geometry_calculator.get_road_length(current_elementary_segment)

            if current_segment_length > self.__max_seg_length_in_meters:
                raise Exception("Max segment reached.")

            # segment reached its required length
            if (current_segment_length >= self.__seg_length_in_meters) or (i == len(road_points) - 1):
                current_segment_length = 0
                end = i
                segment_indexes.append((start, end))
                start = i  # set start index for the next segment

        return segment_indexes


if __name__ == "__main__":
    import unittest

    class ParameterizedUniformStrategyTest(unittest.TestCase):
        def test_straight_only(self):
            strategy = ParameterizedUniformStrategy("2", 0.05)

            road_points = [(x, 0) for x in range(1000)]

            segments = strategy.extract_segments(road_points)

            expected_segments = [(x * 10, x * 10 + 10) for x in range(99)]
            expected_segments.append((990, 999))
            self.assertEqual(segments, expected_segments, "Segment indexes are wrong.")

        def test_road_is_too_short(self):
            strategy = ParameterizedUniformStrategy("2", 0.0005)

            road_points = [(x, 0) for x in range(1000)]

            with self.assertRaises(Exception):
                strategy.extract_segments(road_points)

    unittest.main()



import logging
import abc

from sdc_scissor.obstacle_api.bump import Bump
from sdc_scissor.obstacle_api.delineator import Delineator


class ObstacleFactory(abc.ABC):
    @abc.abstractmethod
    def create_bump(self) -> Bump:
        pass

    @abc.abstractmethod
    def create_delineator(self) -> Delineator:
        pass


if __name__ == "__main__":
    logging.info("abstract_obstacle_factory.py")



import time
import logging

from sdc_scissor.testing_api.test import Test
from sdc_scissor.testing_api.test_loader import TestLoader
from sdc_scissor.testing_api.test_monitor import TestMonitor
from sdc_scissor.testing_api.road_model import RoadModel
from sdc_scissor.simulator_api.abstract_simulator import AbstractSimulator
from sdc_scissor.obstacle_api.obstacle_factory import ObstacleFactory


def _define_obstacles(road_model, obstacle_factory, bump_dist, delineator_dist) -> list:
    logging.info("__define_obstacles")
    obstacles_lst = []
    if obstacle_factory is None:
        return obstacles_lst
    length = int(road_model.ideal_trajectory.length)

    if bump_dist:
        for current_distance in range(bump_dist, length, bump_dist):
            point = road_model.ideal_trajectory.interpolate(-current_distance)
            bump = obstacle_factory.create_bump()
            bump.x_pos = point.x
            bump.y_pos = point.y
            bump.z_pos = -28.0
            obstacles_lst.append(bump)

    if delineator_dist:
        for current_distance in range(delineator_dist, length, delineator_dist):
            point = road_model.center_line.interpolate(-current_distance)
            delineator = obstacle_factory.create_delineator()
            delineator.x_pos = point.x
            delineator.y_pos = point.y
            delineator.z_pos = -28.0
            obstacles_lst.append(delineator)

    return obstacles_lst


class TestRunner:
    def __init__(self, **kwargs) -> None:
        """

        :param kwargs:
        """
        self.test_loader: TestLoader = kwargs.get("test_loader", None)
        self.ml_component = kwargs.get("machine_learning_api", None)
        self.simulator: AbstractSimulator = kwargs.get("simulator", None)
        self.oob: float = kwargs.get("oob", None)
        self.interrupt: bool = kwargs.get("interrupt", None)
        self.bump_dist = kwargs.get("bump_dist", None)
        self.delineator_dist = kwargs.get("delineator_dist", None)
        self.obstacle_factory: ObstacleFactory = kwargs.get("obstacle_factory", None)
        self.fov: list = kwargs.get("fov", None)

    def run_test_suite(self):
        """ """
        logging.info("* run_test_suite")
        self.simulator.open()
        has_execution_failed = False
        test = None
        test_filename = None
        while self.test_loader.has_next() or has_execution_failed:
            if has_execution_failed:
                self.simulator.create_new_instance()
                self.simulator.open()
            else:
                test, test_filename = self.test_loader.next()
            try:
                self.run(test)
                test.save_as_json(file_path=test_filename)
                has_execution_failed = False
                time.sleep(5)
            except Exception:
                # TODO: create a counter to limit the number of trials for a single test
                logging.warning("Test case execution raised an exception!")
                has_execution_failed = True
                test.test_outcome = "ERROR"

        self.simulator.close()

    def run(self, test: Test) -> None:
        """
        Runs the test with the simulator given by instantiation of the test runner.

        :param test: Test object that needs to be executed in simulation
        """
        logging.info("* run")
        time.sleep(5)

        road_model = RoadModel(test.interpolated_road_points)
        obstacles = _define_obstacles(road_model, self.obstacle_factory, self.bump_dist, self.delineator_dist)
        self.simulator.load_scenario(test, obstacles=obstacles)

        # ensure connectivity by blocking the python process for some seconds
        time.sleep(5)

        test_monitor = TestMonitor(self.simulator, test, oob=self.oob, road_model=road_model)
        test_monitor.start_timer()
        self.simulator.start_scenario()

        while not test_monitor.is_test_finished:
            test_monitor.check(self.interrupt)
            time.sleep(0.1)

        test_monitor.stop_timer()
        test_monitor.dump_data()
        self.simulator.stop_scenario()


if __name__ == "__main__":
    logging.info("* test_runner.py")



import logging

from beamngpy import ProceduralCylinder

from sdc_scissor.obstacle_api.delineator import Delineator


class BeamngDelineator(Delineator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self):
        return ProceduralCylinder(
            name="pyCylinder",
            pos=(self.x_pos, self.y_pos, self.z_pos),
            rot=self.rot,
            rot_quat=self.rot_quat,
            radius=self.radius,
            height=self.height,
        )


if __name__ == "__main__":
    logging.info("beamng_delineator.py")



import numpy as np
import logging as log
import random
from time import sleep

from sdc_scissor.testing_api.test_generators.frenetic.src.generators.base_frenet_generator import BaseFrenetGenerator


class CustomFrenetGenerator(BaseFrenetGenerator):
    """
    Generates tests using the frenet framework to determine curvatures.
    """

    def __init__(
        self,
        time_budget=None,
        executor=None,
        map_size=None,
        kill_ancestors=1,
        strict_father=True,
        random_budget=3600,
        crossover_candidates=20,
        crossover_frequency=0,
        count=None,
    ):
        self.count = count
        # Spending 20% of the time on random generation
        # Set this value to 1.0 to generate fully random results.
        self.random_gen_budget = random_budget
        # Margin size w.r.t the map
        self.margin = 10
        # Storing the ancestors of a test that failed to reduce close relatives.
        self.ancestors_of_failed_tests = set()
        self.kill_ancestors = kill_ancestors
        # Only considering tests with a min_oob_distance < threshold for mutation
        # define min_oobd_threshold = 2.0 to remove this feature
        # TODO: Consider updating this value after the initial population
        # df[df.outcome != 'INVALID'].min_oob_distance.quantile(0.25)
        self.min_oobd_threshold = -0.5
        # Set crossover frequency to 0 for no crossover
        self.crossover_candidates = crossover_candidates
        self.crossover_frequency = crossover_frequency

        # Fix number or fix distance policy
        self.max_length = 30
        self.min_step_size = 7
        if map_size < 150:
            # Fix Points
            self.number_of_points = 15
            self.frenet_step = max(self.min_step_size, map_size // self.number_of_points)
        else:
            # Fix Distance
            # Number of generated kappa points depends on the size of the map + random variation
            self.frenet_step = 10
            self.number_of_points = min(map_size // self.frenet_step, self.max_length)

        super().__init__(time_budget=time_budget, executor=executor, map_size=map_size, strict_father=strict_father)

    def start(self):
        log.info("Test generation frenetic.")
        return self.generate_initial_population()
        # self.generate_mutants()
        # sleep(10)

    def generate_initial_population(self):
        # while self.executor.get_remaining_time() > (self.time_budget - self.random_gen_budget):
        road_points_collection = []
        cnt = 0
        while cnt < self.count:
            # print(cnt)
            cnt += 1
            log.info("Random generation. Remaining time %s", 10)
            kappas = self.generate_random_test()
            road_points = self.execute_frenet_test(kappas, frenet_step=self.frenet_step)
            road_points_collection.append(road_points)
            # print(road_points)
        return road_points_collection

    def generate_mutants(self):
        # Iterating the tests according to the value of the min_oob_distance (closer to fail).
        self.recent_count = 0
        while self.executor.get_remaining_time() > 0:
            if 0.0 in set(self.df["visited"]) or 1.0 in set(self.df["visited"]):
                # TODO: The values are become float if there is a nan due to ERROR.
                log.info("Converting visited column to boolean...")
                self.df["visited"] = self.df["visited"].map(lambda x: True if x == 1.0 else False)
            parent = (
                self.df[
                    ((self.df.outcome == "PASS") | (self.df.outcome == "FAIL"))
                    & (~self.df.visited)
                    & (self.df.min_oob_distance < self.min_oobd_threshold)
                ]
                .sort_values("min_oob_distance", ascending=True)
                .head(1)
            )
            if len(parent):
                self.df.at[parent.index, "visited"] = True
                self.mutate_test(parent)
            else:
                # If there is no good parent try random generation
                log.info("There is no good candidate for mutation.")
                kappas = self.generate_random_test()
                self.execute_frenet_test(kappas, frenet_step=self.frenet_step)
            if 0 < self.crossover_frequency <= self.recent_count:
                log.info("Entering crossover phase.")
                self.crossover()
                self.recent_count = 0

    def mutate_test(self, parent):
        # Parent info to be added to the dataframe
        ancestors = []
        if parent.method.item() != "random":
            ancestors = parent.ancestors.item()
        ancestors += [parent.index.item()]

        parent_info = {
            "parent_index": parent.index.item(),
            "parent_outcome": parent.outcome.item(),
            "parent_min_oob_distance": parent.min_oob_distance.item(),
            "generation": parent.generation.item() + 1,
            "ancestors": ancestors,
        }
        if self.kill_ancestors > 0:
            # Looking to close relatives to avoid too similar tests
            for ancestor_id in ancestors[-self.kill_ancestors :]:
                if ancestor_id in self.ancestors_of_failed_tests:
                    return
        # Applying different mutations depending on the outcome
        if parent.outcome.item() == "FAIL":
            self.mutate_failed_test(parent, parent_info)
        else:
            self.mutate_passed_test(parent, parent_info)

    def mutate_passed_test(self, parent, parent_info):
        kappa_mutations = [
            ("add 1 to 5 kappas at the end", self.add_kappas),
            ("randomly remove 1 to 5 kappas", self.randomly_remove_kappas),
            ("remove 1 to 5 kappas from front", lambda ks: ks[random.randint(1, 5) :]),
            ("remove 1 to 5 kappas from tail", lambda ks: ks[: -random.randint(1, 5)]),
            ("increase all kappas 10~20%", self.increase_kappas),
            ("randomly modify 1 to 5 kappas", self.random_modification),
        ]

        self.perform_kappa_mutations(kappa_mutations, parent, parent_info)

    def mutate_failed_test(self, parent, parent_info):
        # Only reversing roads that produced a failure already
        # Mutations to the road
        # TODO: Obtain the kappa values given the cartesians.
        road_points = parent.road.item()
        # Execute reversed original test
        log.info("Mutation function: {:s}".format("reverse road"))
        log.info(
            "Parent ({:s}) {:0.3f} accum_neg_oob and {:0.3f} min oob distance".format(
                parent.outcome.item(), parent.accum_neg_oob.item(), parent.min_oob_distance.item()
            )
        )
        # Do not revisit a reverse road
        self.execute_test(
            road_points[::-1], method="reversed road", extra_info={"visited": True}, parent_info=parent_info
        )

        # mutations that we may want to avoid in tests that passed because they are easily reversible
        kappa_mutations = [
            ("reverse kappas", lambda ks: ks[::-1]),
            ("split and swap kappas", lambda ks: ks[int(len(ks) / 2) :] + ks[: int(len(ks) / 2)]),
            ("flip sign kappas", lambda ks: list(map(lambda x: x * -1.0, ks))),
        ]

        self.perform_kappa_mutations(kappa_mutations, parent, parent_info, extra_info={"visited": True})

    def crossover(self):
        # TODO: Add parent information
        candidates = (
            self.df[
                ((self.df.outcome == "PASS") | (self.df.outcome == "FAIL"))
                & (~self.df.kappas.isna())
                & (self.df.min_oob_distance < self.min_oobd_threshold)
            ]
            .sort_values("min_oob_distance", ascending=True)
            .head(self.crossover_candidates)
        )

        if len(candidates) > 4:
            kids_count = 0
            while self.executor.get_remaining_time() > 0 and kids_count < len(candidates):
                his_id, her_id = random.sample(list(candidates.index), 2)
                father = self.df.iloc[his_id]["kappas"]
                mother = self.df.iloc[her_id]["kappas"]
                if random.random() < 0.5:
                    kids = self.chromosome_crossover(father, mother)
                    name = "chromosome crossover"
                else:
                    kids = self.single_point_crossover(father, mother)
                    name = "single point crossover"
                while self.executor.get_remaining_time() > 0 and len(kids) > 0:
                    kappas = kids.pop()
                    kids_count += 1
                    self.execute_frenet_test(
                        kappas, method=name, frenet_step=self.frenet_step, parent_info={}, extra_info={}
                    )

    @staticmethod
    def chromosome_crossover(him, her):
        """
        him: list of kappas
        her: list of kappas
        returns: a list of kappas of the length of the shortest list
        """
        son = []
        for i in range(min(len(him), len(her))):
            if random.random() < 0.5:
                son.append(him[i])
            else:
                son.append(her[i])
        return [son]

    @staticmethod
    def single_point_crossover(him, her):
        """
        him: list of kappas
        her: list of kappas
        returns: Two lists of kappas
        """
        son = him[len(him) // 2 :] + her[: len(her) // 2]
        daughter = her[len(her) // 2 :] + him[: len(him) // 2]
        return [son, daughter]

    def perform_kappa_mutations(self, kappa_mutations, parent, parent_info, extra_info={}):
        # Only considering paths with more than 10 kappa points for mutations
        # kappas might be empty if the parent was obtained from reverse road points mutation
        kappas = parent.kappas.item()
        if kappas and len(kappas) > 10:
            i = 0
            while self.executor.get_remaining_time() > 0 and i < len(kappa_mutations):
                name, function = kappa_mutations[i]
                log.info("Generating mutants. Remaining time %s", self.executor.get_remaining_time())
                log.info("Mutation function: {:s}".format(name))
                log.info(
                    "Parent ({:s}) {:0.3f} accum_neg_oob and {:0.3f} min oob distance".format(
                        parent.outcome.item(), parent.accum_neg_oob.item(), parent.min_oob_distance.item()
                    )
                )
                m_kappas = function(kappas)
                outcome, _ = self.execute_frenet_test(
                    m_kappas, frenet_step=self.frenet_step, method=name, parent_info=parent_info, extra_info=extra_info
                )

                # When there is a mutant of this branch that fails, we stop mutating this branch.
                if outcome == "FAIL" and self.kill_ancestors > 0:
                    for ancestor_id in parent_info["ancestors"][-self.kill_ancestors :]:
                        self.ancestors_of_failed_tests.add(ancestor_id)
                    break
                i += 1

    @staticmethod
    def get_next_kappa(last_kappa, kappa_bound=0.05, kappa_delta=0.07):
        return random.choice(
            np.linspace(max(-kappa_bound, last_kappa - kappa_delta), min(kappa_bound, last_kappa + kappa_delta))
        )

    @staticmethod
    def increase_kappas(kappas):
        m = 1.1 + 0.1 * random.random()
        return list(map(lambda x: x * m, kappas))

    @staticmethod
    def add_kappas(kappas):
        # number of kappas to added
        k = random.randint(1, 5)
        modified_kappas = kappas[:]
        last_kappa = kappas[-1]
        while k > 0:
            # Randomly add a kappa
            modified_kappas.append(CustomFrenetGenerator.get_next_kappa(last_kappa))
            k -= 1
        return modified_kappas

    @staticmethod
    def randomly_remove_kappas(kappas):
        # number of kappas to be removed
        k = random.randint(1, 5)
        modified_kappas = kappas[:]
        while k > 0 and len(modified_kappas) > 5:
            # Randomly remove a kappa
            i = random.randint(0, len(modified_kappas) - 1)
            del modified_kappas[i]
            k -= 1
        return modified_kappas

    @staticmethod
    def random_modification(kappas):
        # number of kappas to be modified
        k = random.randint(1, 5)
        # Randomly modified kappa
        indexes = random.sample(range(len(kappas) - 1), k)
        modified_kappas = kappas[:]
        for i in indexes:
            modified_kappas[i] += random.choice(np.linspace(-0.05, 0.05))
        return modified_kappas

    def generate_random_test(self, kappa_delta=0.05, kappa_bound=0.07):
        """Generates a test using frenet framework to determine the curvature of the points.
         Currently using an initial setup similar to the GUI.
         TODO: Make the frenet setup part of the experiment to adapt w.r.t. the output of the tests.
        Args:
            frenet_step: The distance between to points.
            theta0: The initial angle of the line. (1.57 == 90 degrees)
            kappa_delta: The maximum difference between two consecutive kappa values.
            kappa_bound: The maximum value of kappa allowed.
        Returns:
            a list of kappa values and its cartesian representation.
        """
        points = self.number_of_points + random.randint(-5, 5)
        # Producing randomly generated kappas for the given setting.
        kappas = [0.0] * points
        for i in range(len(kappas)):
            kappas[i] = CustomFrenetGenerator.get_next_kappa(kappas[i - 1], kappa_bound, kappa_delta)

        return kappas


class Frenetic(CustomFrenetGenerator):
    def __init__(self, time_budget=None, executor=None, map_size=None):
        super().__init__(
            time_budget=time_budget,
            executor=executor,
            map_size=map_size,
            kill_ancestors=0,
            strict_father=False,
            random_budget=3600,
            crossover_candidates=20,
            crossover_frequency=40,
        )



#!/usr/bin/env python
"""
Prioritize tests using NSGA-II
"""


import csv
from genericpath import exists
import sys
import os.path as path
import os
import numpy as np
import math


from crossover.PMX import PMXCrossover
from mutation.HybridMut import HybridMut
from problem.TestPrioritizationMultiObjectiveProblem import TestPrioritizationMultiObjectiveProblem

# pymoo libs
from pymoo.algorithms.moo.nsga2 import NSGA2

from pymoo.factory import get_sampling
from pymoo.optimize import minimize
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.visualization.scatter import Scatter

# video recorder
from pyrecorder.recorder import Recorder
from pyrecorder.writers.video import Video

# sklearn
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler


MINIMUM_USER_INPTUS = 3
MAXIMUM_USER_INPTUS = 5


def fault_detection(X, lables, costs):
    length = len(lables)
    faults = np.array(lables)
    faults = np.where(faults == "safe", 0, faults)
    faults = np.where(faults == "unsafe", 1, faults).astype(float)

    x_values = np.zeros(length)
    y_values = np.zeros(length)

    total_faults = 0.0
    total_costs = 0.0

    for index in range(0, length):
        total_costs += costs[X[index]]
        total_faults += faults[X[index]]

        x_values[index] = total_costs
        y_values[index] = total_faults

    return x_values, y_values


def get_APFD(X, lables, costs):
    a, b = fault_detection(X, lables, costs)

    APFD = np.trapz(b, a) / (np.max(a) * np.max(b))
    print(f"APFD is {APFD}")
    return APFD


def get_costs_and_features(csv_file_path):
    costs = []
    features = []
    lables = []
    with open(csv_file_path) as csv_file:
        csv_reader = csv.reader(csv_file)
        next(csv_reader)
        for row in csv_reader:
            costs.append(row[18])
            lables.append(row[19])
            features.append(row[:16])

    costs = np.array(costs).astype(float)
    features = np.array(features).astype(float)

    return costs, features, lables


def get_utopia(fitness_values):
    best_f1 = float("inf")
    best_f2 = float("inf")
    for fc in fitness_values:
        if fc[0] < best_f1:
            best_f1 = fc[0]
        if fc[1] < best_f2:
            best_f2 = fc[1]
    return [best_f1, best_f2]


def validate_user_input(user_input):
    # Check number of args
    if len(user_input) < MINIMUM_USER_INPTUS or len(user_input) > MAXIMUM_USER_INPTUS:
        print_help()
        exit()

    # Validate Features CSV file
    csv_file_path = user_input[0]

    if not path.exists(csv_file_path):
        print_help()
        exit()

    # Check configuration name
    config_name = user_input[1]
    if config_name == "":
        print_help()
        exit()

    output_dir = user_input[2]
    if not os.path.isdir(output_dir):
        print(f"{output_dir} is not a directory. Creating  {output_dir}")
        os.mkdir(output_dir)

    # check population size
    population_size = 100
    if len(user_input) >= 4:
        population_size = int(user_input[3])

    budget = 2000
    if len(user_input) == 5:
        budget = int(user_input[4])

    return csv_file_path, config_name, output_dir, population_size, budget


def print_help():
    """
    Print help text for wrong user input.
    """
    print(f"Expected at least {MINIMUM_USER_INPTUS} input arguments and at most {MAXIMUM_USER_INPTUS}.")
    print("arg 1 = Features CSV file path")
    print("arg 2 = Configuration name")
    print("arg 3 = Output directory")
    print("arg 4 (optional) = poulation size")
    print("arg 5 (optional) = number of generations")


def save_plot(output_dir, problem, res):
    plot = Scatter()
    plot.add(problem.pareto_front(), plot_type="line", color="black", alpha=0.7)
    plot.add(res.F, facecolor="none", edgecolor="red")
    plot.save(path.join(output_dir, "plot.png"))


class MyDisplay(MultiObjectiveDisplay):
    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)


def get_algorithm(algorithm, sampling_var, population_size=100):
    if algorithm == "NSGAII":
        return NSGA2(
            pop_size=population_size,
            sampling=sampling_var,
            crossover=PMXCrossover(),
            mutation=HybridMut(),
            eliminate_duplicates=True,
        )
    else:
        raise Exception("Unknown GA!")


def select_single_solution(res, scaler):
    normalized_solutions = scaler.fit_transform(res.F)
    print(normalized_solutions)

    best_solution = 0
    minimum_distance = float("inf")
    for index, solution_fitnesses in enumerate(normalized_solutions):
        f1 = solution_fitnesses[0]
        f2 = solution_fitnesses[1]
        distance = math.sqrt(f1**2 + f2**2)
        if distance < minimum_distance:
            minimum_distance = distance
            best_solution = index

    return best_solution


def save_solution(output_dir, solution):
    final_solution = solution.tolist()
    # print(final_solution)
    file_dir = os.path.join(output_dir, "solution.txt")
    # np.savetxt(file_dir, final_solution, fmt=str)
    textfile = open(file_dir, "w")
    for element in final_solution:
        textfile.write(str(element) + ", ")
    textfile.close


def main(user_input: list = None):
    cwd = os.getcwd()
    scaler = MinMaxScaler()

    """
    Main entry point when run the search process

    Args:
        user_input: argv[1]: Features CSV file path

    """
    # 1- Validating user inputs
    (csv_file_path, config_name, output_dir, population_size, budget) = validate_user_input(user_input)
    output_dir = os.path.join(output_dir, config_name)
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    global costs
    # 2- Read CSV file
    costs, features, lables = get_costs_and_features(csv_file_path)
    genome_length = len(costs)

    # 3- Normalize all the features (min-max normalization)
    features = scaler.fit_transform(features)

    # 4- PCA
    pca = PCA()
    pca.fit(features)
    features = pca.transform(features)

    # 5- Compute the pairwse (Euclidean) distance between test cases
    global distances
    distances = np.sqrt(((features[:, :, None] - features[:, :, None].T) ** 2).sum(1))

    # 6- Run GA
    ## problem
    problem = TestPrioritizationMultiObjectiveProblem(genome_length, distances, costs)

    ## termination
    # termination = MultiObjectiveDefaultTermination(n_last=50, f_tol = 0)

    ## initialization (random solutions)
    sampling_var = get_sampling("perm_random")

    ## algorithm
    algorithm = get_algorithm("NSGAII", sampling_var, population_size)

    ## Start the search process
    res = minimize(problem, algorithm, ("n_gen", budget), save_history=False, display=MyDisplay(), verbose=True)

    ## 7 - print and save the outcome of the search process
    save_plot(output_dir, problem, res)

    # # use the video writer as a resource
    # with Recorder(Video(path.join(data_path,"ga.mp4"))) as rec:

    # # for each algorithm object in the history
    #     for entry in res.history:
    #         sc = Scatter(title=("Gen %s" % entry.n_gen))
    #         sc.add(entry.pop.get("F"))
    #         sc.add(entry.problem.pareto_front(), plot_type="line", color="black", alpha=0.7)
    #         sc.do()

    #         # finally record the current visualization to the video
    #         rec.record()

    # 8 - Select a solution
    best_solution_index = select_single_solution(res, scaler)

    print(f"Best solution has fitness values  {res.F[best_solution_index]}")

    final_solution = res.X[best_solution_index]
    final_solution_fitness = res.F[best_solution_index]

    # 9 - Save selected solution
    print(f"final_solution {final_solution.shape}")
    save_solution(output_dir, final_solution)

    # 9 - Calculate APFD
    get_APFD(final_solution, lables, costs)


if __name__ == "__main__":
    main(sys.argv[1:])



import numpy as np


def frenetv_to_cartesian(x0, y0, theta0, ss, kappas):
    """
    Trapezoidal integration to compute Cartesian coordinates from given curvature values.
    """
    xs = np.zeros(len(kappas))
    ys = np.zeros(len(kappas))
    thetas = np.zeros(len(kappas))
    xs[0] = x0
    ys[0] = y0
    thetas[0] = theta0
    for i in range(thetas.shape[0] - 1):
        ss_diff_half = (ss[i + 1] - ss[i]) / 2.0
        thetas[i + 1] = thetas[i] + (kappas[i + 1] + kappas[i]) * ss_diff_half
        xs[i + 1] = xs[i] + (np.cos(thetas[i + 1]) + np.cos(thetas[i])) * ss_diff_half
        ys[i + 1] = ys[i] + (np.sin(thetas[i + 1]) + np.sin(thetas[i])) * ss_diff_half
    return (xs, ys)



import numpy as np
from pymoo.model.sampling import Sampling
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.solution import Solution

import sdc_scissor.testing_api.test_generators.ambiegen.config as cf
from sdc_scissor.testing_api.test_generators.ambiegen.Utils.road_gen import RoadGen


class GenerateTestCaseSampling(Sampling):

    """
    Module to generate the initial population
    """

    def _do(self, problem, n_samples, **kwargs):
        generator = RoadGen(
            cf.model["map_size"], cf.model["min_len"], cf.model["max_len"], cf.model["min_angle"], cf.model["max_angle"]
        )
        X = np.full((n_samples, 1), None, dtype=np.object)

        for i in range(n_samples):
            states = generator.test_case_generate()
            s = Solution()
            s.states = states
            X[i, 0] = s
        return X



import numpy as np
import math as m
from shapely.geometry import Point

from scipy.interpolate import splprep, splev
from shapely.geometry import LineString, Point
from numpy.ma import arange


class Car:
    """Class that conducts transformations to vectors automatically,
    using the commads "go straight", "turn left", "turn right".
    As a result it produces a set of points corresponding to a road
    """

    def __init__(self, speed, steer_ang, map_size):
        self.speed = speed
        self.map_size = map_size
        self.str_ang = steer_ang
        self.str_ang_o = steer_ang

    def interpolate_road(self, road):

        test_road = LineString([(t[0], t[1]) for t in road])

        length = test_road.length

        num_nodes = int(length)
        if num_nodes < 20:
            num_nodes = 20

        old_x_vals = [t[0] for t in road]
        old_y_vals = [t[1] for t in road]

        if len(old_x_vals) == 2:
            k = 1
        elif len(old_x_vals) == 3:
            k = 2
        else:
            k = 3
        f2, u = splprep([old_x_vals, old_y_vals], s=0, k=k)
        # step_size = 1 / (length) * 8

        step_size = 1 / num_nodes

        xnew = arange(0, 1 + step_size, step_size)

        x2, y2 = splev(xnew, f2)

        nodes = list(zip(x2, y2))

        return nodes

    def get_distance(self, road, x, y):
        p = Point(x, y)
        return p.distance(road)

    def go_straight(self):
        self.x = self.speed * np.cos(m.radians(self.angle)) / 2.3 + self.x
        self.y = self.speed * np.sin(m.radians(self.angle)) / 2.3 + self.y
        self.tot_x.append(self.x)
        self.tot_y.append(self.y)
        return

    def turn_right(self):

        self.str_ang = m.degrees(m.atan(1 / self.speed * 2 * self.distance))
        self.angle = -self.str_ang + self.angle
        self.x = self.speed * np.cos(m.radians(self.angle)) / 3 + self.x
        self.y = self.speed * np.sin(m.radians(self.angle)) / 3 + self.y
        self.tot_x.append(self.x)
        self.tot_y.append(self.y)
        return

    def turn_left(self):
        self.str_ang = m.degrees(m.atan(1 / self.speed * 2 * self.distance))
        self.angle = self.str_ang + self.angle
        self.x = self.speed * np.cos(m.radians(self.angle)) / 3 + self.x
        self.y = self.speed * np.sin(m.radians(self.angle)) / 3 + self.y

        self.tot_x.append(self.x)
        self.tot_y.append(self.y)

        return

    def get_angle(self, node_a, node_b):
        vector = np.array(node_b) - np.array(node_a)
        cos = vector[0] / (np.linalg.norm(vector))

        angle = m.degrees(m.acos(cos))

        if node_a[1] > node_b[1]:
            return -angle
        else:
            return angle

    def execute_road(self, nodes):

        self.x = 0
        self.y = 0

        old_x_vals = [t[0] for t in nodes]
        old_y_vals = [t[1] for t in nodes]

        self.road_x = old_x_vals
        self.road_y = old_y_vals

        self.angle = 0
        self.tot_x = []
        self.tot_y = []
        self.tot_dist = []
        self.final_dist = []
        self.distance = 0

        road = LineString([(t[0], t[1]) for t in nodes])
        mini_nodes1 = nodes[: round(len(nodes) / 2)]
        mini_nodes2 = nodes[round(len(nodes) / 2) :]
        if (len(mini_nodes1) < 2) or (len(mini_nodes2) < 2):
            return 0, []
        mini_road1 = LineString([(t[0], t[1]) for t in mini_nodes1])
        mini_road2 = LineString([(t[0], t[1]) for t in mini_nodes2])
        road_split = [mini_road1, mini_road2]

        if (road.is_simple is False) or (is_too_sharp(_interpolate(nodes)) is True):
            fitness = 0
        else:
            init_pos = nodes[0]
            self.x = init_pos[0]
            self.y = init_pos[1]

            self.angle = self.get_angle(nodes[0], nodes[1])

            self.tot_x.append(self.x)
            self.tot_y.append(self.y)

            i = 0

            for p, mini_road in enumerate(road_split):

                current_length = 0
                if p == 1:

                    self.x = mini_nodes2[0][0]
                    self.y = mini_nodes2[0][1]
                    self.angle = self.get_angle(mini_nodes1[-1], mini_nodes2[0])

                current_pos = [(self.x, self.y)]

                while (current_length < mini_road.length) and i < 1000:
                    distance = self.get_distance(mini_road, self.x, self.y)
                    self.distance = distance

                    self.tot_dist.append(distance)
                    if distance <= 1:
                        self.go_straight()
                        current_pos.append((self.x, self.y))
                        self.speed += 0.3

                    else:
                        angle = -1 + self.angle
                        x = self.speed * np.cos(m.radians(angle)) + self.x
                        y = self.speed * np.sin(m.radians(angle)) + self.y

                        distance_right = self.get_distance(mini_road, x, y)

                        angle = 1 + self.angle
                        x = self.speed * np.cos(m.radians(angle)) + self.x
                        y = self.speed * np.sin(m.radians(angle)) + self.y

                        distance_left = self.get_distance(mini_road, x, y)

                        if distance_right < distance_left:
                            self.turn_right()
                            current_pos.append((self.x, self.y))
                        else:
                            self.turn_left()
                            current_pos.append((self.x, self.y))

                        self.speed -= 0.1

                    current_road = LineString(current_pos)
                    current_length = current_road.length

                    i += 1

            fitness = max(self.tot_dist) * (-1)

            car_path = LineString(zip(self.tot_x, self.tot_y))
            if car_path.is_simple == False:
                fitness = 0

        return fitness, [self.tot_x, self.tot_y]


def find_circle(p1, p2, p3):
    """
    Returns the center and radius of the circle passing the given 3 points.
    In case the 3 points form a line, returns (None, infinity).
    """
    temp = p2[0] * p2[0] + p2[1] * p2[1]
    bc = (p1[0] * p1[0] + p1[1] * p1[1] - temp) / 2
    cd = (temp - p3[0] * p3[0] - p3[1] * p3[1]) / 2
    det = (p1[0] - p2[0]) * (p2[1] - p3[1]) - (p2[0] - p3[0]) * (p1[1] - p2[1])

    if abs(det) < 1.0e-6:
        return np.inf

    # Center of circle
    cx = (bc * (p2[1] - p3[1]) - cd * (p1[1] - p2[1])) / det
    cy = ((p1[0] - p2[0]) * cd - (p2[0] - p3[0]) * bc) / det

    radius = np.sqrt((cx - p1[0]) ** 2 + (cy - p1[1]) ** 2)
    # print(radius)
    return radius


def min_radius(x, w=5):
    mr = np.inf
    nodes = x
    for i in range(len(nodes) - w):
        p1 = nodes[i]
        p2 = nodes[i + int((w - 1) / 2)]
        p3 = nodes[i + (w - 1)]
        radius = find_circle(p1, p2, p3)
        if radius < mr:
            mr = radius
    if mr == np.inf:
        mr = 0

    return mr * 3.280839895  # , mincurv


def _interpolate(the_test):
    """
    Interpolate the road points using cubic splines and ensure we handle 4F tuples for compatibility
    """
    rounding_precision = 3
    interpolation_distance = 1
    smoothness = 0
    min_num_nodes = 20

    old_x_vals = [t[0] for t in the_test]
    old_y_vals = [t[1] for t in the_test]

    # This is an approximation based on whatever input is given
    test_road_lenght = LineString([(t[0], t[1]) for t in the_test]).length
    num_nodes = int(test_road_lenght / interpolation_distance)
    if num_nodes < min_num_nodes:
        num_nodes = min_num_nodes

    assert len(old_x_vals) >= 2, "You need at leas two road points to define a road"
    assert len(old_y_vals) >= 2, "You need at leas two road points to define a road"

    if len(old_x_vals) == 2:
        # With two points the only option is a straight segment
        k = 1
    elif len(old_x_vals) == 3:
        # With three points we use an arc, using linear interpolation will result in invalid road tests
        k = 2
    else:
        # Otheriwse, use cubic splines
        k = 3

    pos_tck, pos_u = splprep([old_x_vals, old_y_vals], s=smoothness, k=k)

    step_size = 1 / num_nodes
    unew = arange(0, 1 + step_size, step_size)

    new_x_vals, new_y_vals = splev(unew, pos_tck)

    # Return the 4-tuple with default z and defatul road width
    return list(
        zip(
            [round(v, rounding_precision) for v in new_x_vals],
            [round(v, rounding_precision) for v in new_y_vals],
            [-28.0 for v in new_x_vals],
            [8.0 for v in new_x_vals],
        )
    )


def is_too_sharp(the_test, TSHD_RADIUS=47):
    if TSHD_RADIUS > min_radius(the_test) > 0.0:
        check = True
        # print("TOO SHARP")
    else:
        check = False
    # print(check)
    return check



import numpy as np
import random


from pymoo.core.mutation import Mutation


class HybridMut(Mutation):
    def __init__(self, prob=1.0):
        super().__init__()
        self.prob = prob

    def _do(self, problem, X, **kwargs):
        X = X[: int(len(X) / 2)]
        Y = X.copy()

        for child in Y:
            mutation_points = []
            while len(mutation_points) < 2:
                temp_point = random.randint(0, len(child) - 1)
                if temp_point not in mutation_points:
                    mutation_points.append(temp_point)
            mutation_selector_prob = random.uniform(0, 1)
            if mutation_selector_prob <= 0.33:

                temp = child[mutation_points[0]]
                child[mutation_points[0]] = child[mutation_points[1]]
                child[mutation_points[1]] = temp

            elif mutation_selector_prob <= 0.66:

                min_index = min(mutation_points)
                max_index = max(mutation_points)

                child[min_index : max_index + 1] = np.array(list(reversed(child[min_index : max_index + 1])))

            else:

                temp = child[mutation_points[0]]
                child = np.delete(child, mutation_points[0])
                child = np.insert(child, mutation_points[1], temp)

        return Y



import logging
import math
import time

from beamngpy import BeamNGpy, Scenario, Road, Vehicle
from beamngpy.sensors import Electrics
from shapely.geometry import LineString

from sdc_scissor.simulator_api.abstract_simulator import AbstractSimulator
from sdc_scissor.testing_api.test import Test


class BeamNGSimulator(AbstractSimulator):
    """
    This class implements the interface for the specific BeamNG.tech simulator.
    """

    def __init__(self, host: str, port: int, home: str, user: str, rf: float, max_speed: float, fov: int):
        """
        API for enabling inter-process communication with the BeamNG simulator.

        :param host: Hostname of the machine, e.g., 'localhost'
        :param port: The port for communication with the BeamNG process
        :param home: Path to BeamNG installation
        :param user: The user path (path to your license key file)
        :param rf: The risk factor, e.g., 1.5
        :param max_speed: The maximal speed allowed for a vehicle in km/h
        :param fov: The field of view  for a Camera e.g., 120
        """
        super().__init__()
        self.host = host
        self.port = port
        self.home = home
        self.user = user
        self.beamng = BeamNGpy(host, port, home=home, user=user)
        self.vehicle = None
        self.car_state = None
        self.scenario = None
        self.rf = rf
        self.max_speed = max_speed
        self.fov = fov

    def open(self):
        """
        Start the BeamNG.tech process
        """
        self.beamng.open()

    def close(self):
        """
        Quit the BeamNG.tech process
        """
        self.beamng.close()

    def create_new_instance(self):
        """
        Restart the BeamNG.tech process
        """
        logging.info("restart simulator")
        try:
            self.beamng.close()
        finally:
            logging.info("creating a new BeamNGpy instance")
            time.sleep(5)
            self.beamng = BeamNGpy(self.host, self.port, home=self.home, user=self.user)

    def stop_scenario(self):
        """ """
        logging.info("stop_scenario")
        self.beamng.stop_scenario()

    def start_scenario(self):
        """ """
        logging.info("start_scenario")
        self.beamng.start_scenario()
        self.vehicle.ai_set_mode("span")
        self.vehicle.ai_drive_in_lane(lane=True)
        self.vehicle.ai_set_aggression(self.rf)
        self.vehicle.set_color(rgba=(0, 0, 1, 0.5))
        self.vehicle.ai_set_speed(self.__kmh_to_ms(self.max_speed))
        self.vehicle.ai_set_waypoint("end_point")

    @staticmethod
    def __kmh_to_ms(kmh):
        """

        :param kmh:
        :return:
        """
        return kmh / 3.6

    def load_scenario(self, test: Test, obstacles: list):
        """

        :param test:
        :param obstacles:
        """
        logging.info("load_scenario")
        self.scenario = Scenario("tig", "example")
        road = Road(material="tig_road_rubber_sticky", rid="flat_road", interpolate=True)

        # Ensure not overriding the test object (copy first the whole list)
        road_nodes = test.interpolated_road_points.copy()
        for road_node in road_nodes:
            road_node.extend([-28, 10])

        road.nodes.extend(road_nodes)

        self.scenario.add_road(road)

        logging.info("* generate obstacle points")
        for obstacle in obstacles:
            self.scenario.add_procedural_mesh(obstacle.get())

        self.vehicle = Vehicle(vid="ego_vehicle", model="etk800", licence="Scissor")
        electrics = Electrics()
        self.vehicle.attach_sensor("electrics", electrics)

        x_start, y_start, z_start, x_dir, y_dir, alpha = self.__compute_start_position(road_nodes)
        # TODO: Calculate the start rotation of the car in quaterion
        rot_quat = (0, 0, 1, -0.1)
        rot = (0, 0, alpha / (2 * math.pi))
        self.scenario.add_vehicle(self.vehicle, pos=(x_start, y_start, z_start), rot_quat=rot_quat)

        end_point = road_nodes[-1][:3]

        self.scenario.add_checkpoints(positions=[end_point], scales=[(5, 5, 5)], ids=["end_point"])
        self.scenario.make(self.beamng)
        self.beamng.load_scenario(self.scenario)

    def update_car(self):
        """ """
        logging.info("update_car")
        self.vehicle.update_vehicle()
        _ = self.beamng.poll_sensors(self.vehicle)  # otherwise, the values are not updated (bug of beamngpy)
        self.car_state = self.vehicle.state

    def get_car_position(self):
        """
        Compute the car's position

        :return: x,y,z coordinates of the car
        """
        logging.info("get_car_position")
        x_pos = self.car_state["pos"][0]
        y_pos = self.car_state["pos"][1]
        z_pos = self.car_state["pos"][2]
        return x_pos, y_pos, z_pos

    @staticmethod
    def __compute_start_position(road_nodes):
        """
        Compute start position of the car. The car should be on the right lane in the beginning of the road.

        :param road_nodes: Road nodes specifying the road for a BeamNG scenario
        :return: The coordinates of the start position of the car.
        """
        logging.info("compute_start_position")
        first_road_point = road_nodes[0]
        second_road_point = road_nodes[1]
        x_dir, y_dir = (second_road_point[0] - first_road_point[1], second_road_point[1] - first_road_point[1])
        x_dir_norm, y_dir_norm = (
            x_dir / math.sqrt(x_dir**2 + y_dir**2),
            y_dir / math.sqrt(x_dir**2 + y_dir**2),
        )
        alpha = math.asin(y_dir_norm)

        center_line = LineString(coordinates=[(x[0], x[1]) for x in road_nodes])
        optimal_trajectory = center_line.parallel_offset(distance=2.5)
        start_point = optimal_trajectory.interpolate(distance=-2.5)
        logging.info("start_point: {}".format(start_point))

        start_position = (start_point.x, start_point.y)

        return (start_position[0], start_position[1], first_road_point[2], x_dir, y_dir, alpha)


if __name__ == "__main__":
    logging.info("beamng_simulator.py")



ga = {"population": 100, "n_gen": 75, "mut_rate": 0.4, "cross_rate": 1}

model = {
    "speed": 9,  # parameter for the simplified car model
    "map_size": 200,
    "steer_ang": 12,  # a parameter for the simplified car model
    "min_len": 5,  # minimal possible distance in meters
    "max_len": 30,  # maximal possible disance to go straight in meters
    "min_angle": 10,  # minimal angle of rotation in degrees
    "max_angle": 85,  # maximal angle of rotation in degrees
}

files = {"ga_archive": ".\\GA_archive\\", "tc_img": ".\\TC_img\\", "tc_file": ".\\TC_file\\"}



import logging

import pandas as pd
import numpy as np

from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.model_selection import KFold, cross_validate, train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import LinearSVC


class CostEffectivenessEvaluator:
    def __init__(self, data_frame: pd.DataFrame, label: str, time_attribute: str):
        """

        :param data_frame:
        :param label:
        :param time_attribute:
        """
        self.data_frame = data_frame
        self.label = label
        self.time_attribute = time_attribute
        self.X_model_attributes = [
            "direct_distance",
            "max_angle",
            "max_pivot_off",
            "mean_angle",
            "mean_pivot_off",
            "median_angle",
            "median_pivot_off",
            "min_angle",
            "min_pivot_off",
            "num_l_turns",
            "num_r_turns",
            "num_straights",
            "road_distance",
            "std_angle",
            "std_pivot_off",
            "total_angle",
        ]

    def evaluate(self):
        """ """
        logging.info("evaluate")
        X_attributes = self.X_model_attributes + [self.time_attribute]

        # train models CV
        X = self.data_frame[X_attributes].to_numpy()
        # print(X)
        # print(X[:, -1])
        # TODO: provide preprocessing options to the user???
        # X = preprocessing.normalize(X)
        # X = preprocessing.scale(X)
        y = self.data_frame[self.label].to_numpy()
        y[y == "FAIL"] = 1
        y[y == "PASS"] = 0
        y = np.array(y, dtype="int32")

        X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.7)

        X_test_time = X_test[:, -1]

        X_train = X_train[:, :-1]
        X_test = X_test[:, :-1]

        classifiers = {}

        classifiers = {
            "random_forest": RandomForestClassifier(),
            "gradient_boosting": GradientBoostingClassifier(),
            # 'multinomial_naive_bayes': MultinomialNB(),
            "SVM": LinearSVC(),
            "gaussian_naive_bayes": GaussianNB(),
            "logistic_regression": LogisticRegression(max_iter=10000),
            "decision_tree": DecisionTreeClassifier(),
        }

        for name, estimator in classifiers.items():
            estimator.fit(X_train, y_train)
            y_pred = estimator.predict(X_test)
            nr_unsafe_predicted = np.sum(y_pred)

            rand_sim_times = []
            nr_trials = 10
            for i in range(nr_trials):
                random_unsafe_predicted = np.random.permutation(
                    np.append(
                        np.ones(nr_unsafe_predicted, dtype="int32"),
                        np.zeros(len(y_pred - nr_unsafe_predicted), dtype="int32"),
                    )
                )
                rand_sim_times.append(np.sum(X_test_time[random_unsafe_predicted]))

            random_tot_sim_time = np.mean(rand_sim_times)

            sdc_scissor_tot_sim_time = np.sum(X_test_time[y_pred])
            print("SDC-SCISSOR")
            print("{}:\tnr_tests: {}\ttot_sim_time {}".format(name, nr_unsafe_predicted, sdc_scissor_tot_sim_time))
            print("RANDOM BASELINE:")
            print("nr_tests: {}\ttot_sim_time {}".format(nr_unsafe_predicted, random_tot_sim_time))
            print("random_baseline_time/sdc_scissor_time = {}".format(random_tot_sim_time / sdc_scissor_tot_sim_time))
            print("sdc_scissor_time/random_baseline_time = {}\n".format(sdc_scissor_tot_sim_time / random_tot_sim_time))


if __name__ == "__main__":
    logging.info("cost_effectiveness_evaluator.py")
